# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Information about the environment this is running on 
"""

import sys, os, subprocess
import tempfile, json
import logging
import re
from railroadtracks import core

logger = logging.getLogger(__name__)


class MissingSoftware(ValueError):
    pass


def _is_executable(path):
        return os.path.isfile(path) and os.access(path, os.X_OK)

def _find_executable(executable):
    path, fname = os.path.split(executable)
    if path:
        if _is_executable(executable):
            return executable
    else:
        for env_path in os.environ["PATH"].split(os.pathsep):
            path = env_path.strip('"')
            candidate_path = os.path.join(path, executable)
            if _is_executable(candidate_path):
                return candidate_path
    return None
class Executable(object):

    def __init__(self, executable):
        """ Name of executable, or full path to executable. """
        if executable.startswith('/'):
            path = executable
        else:
            path = _find_executable(executable)
        if (path is None) or (not os.path.exists(path)):
            raise MissingSoftware('%s cannot be found' % executable)
        self.path = path

    def __repr__(self):
        s = [super(Executable, self).__repr__(),
             '  absolute path: %s' % self.path]
        return '\n'.join(s)

    @staticmethod
    def ispresent(name):
        try:
            e = Executable(name)
        except MissingSoftware:
            return False
        return True

class R(Executable):

    __version = None
    rjsonversion = None

    def getversion(self):
        if self.__version is None:
            with core.Popen((self.path, '--version'), \
                            stdout=subprocess.PIPE) as proc:
                version = proc.stdout.readline()
                self.__version = re.sub(b'R version (.+)', b'\\1', version)
        logger.info('R version is %s' % self.__version)
        return self.__version    
    version = property(getversion, None, None, 'R version')


    def packageversion(self, name):
        """ Check whether an R package is installed and return the version number.
        If the package cannot be loaded (generally because it is not installed),
        an exception :class:`MissingSoftware` is raised.

        :param name: name of the package

        """
        code = """res <- suppressMessages(suppressWarnings(require("%s", quietly=TRUE))); cat(res)"""
        cmd = (self.path, '--slave', '--no-restore', '-e', code % name)

        rsays = subprocess.check_output(cmd).rstrip()
        if rsays == b'FALSE':
            raise MissingSoftware("The R package '%s' is not installed" % name)

        code = 'res <- suppressMessages(packageVersion("%s")); cat(as.character(res))'
        cmd = (self.path, '--slave', '-e', code % name)
        with open(os.devnull, "w") as fnull,\
             core.Popen(cmd, stdout=subprocess.PIPE,
                        stderr=fnull) as proc:
            version = proc.stdout.readline().rstrip()
            tmp = proc.stdout.read()

        logger.info('R package "%s" version is %s' % (name, version))
        return version

    def packageversion_or_none(self, name):
        """ Return the package version, or `None` if the R package cannot be loaded.
        This method is a wrapper around the method `packageversion`.
        """
        try:
            version = self.packageversion(name)
        except MissingSoftware:
            version = None
        return version

    def run_snippet(self, code, var_in,
                    magicvariable = 'railroadtracks_import'):
        """
        :param code: snippet of R code as text
        :param var_in: dict of variables to import into R's global environment before the snippet is evaluated.
                       this is a rather ad-hoc system, and significant performance problems will occur
                       with large data, or complex data structures.

        For example:

        >>> code = 'print(x+1)'
        >>> d = {'x': 2}
        >>> r = R()
        >>> r.run_snippet(code, d)
        [1] 3
        """
        if self.rjsonversion is None:
            # package rjson used to pass context data
            self.rjsonversion = self.packageversion('rjson')

        code_json_extract = """
        # paranoid check
        if (exists("%(magicvariable)s", envir=.GlobalEnv)) {
        cat("*** The variable name '%(magicvariable)s' is reserved !")
        print()
        }
        
        railroadtracks_import <- local({
            suppressWarnings(require("rjson", quietly=TRUE))
            con <- file("%(filename)s", "r")
            res <- rjson::fromJSON(file=con)
            close(con)
            res
        })
        """ 

        if logger.getEffectiveLevel() == logging.DEBUG:
            delete_tempfile = False
        else:
            delete_tempfile = True
        with tempfile.NamedTemporaryFile(delete=delete_tempfile,
                                         mode='wt') as fh_out:
            var_in_json = json.dump(var_in, fh_out)
            fh_out.flush()
            if not delete_tempfile:
                logger.debug('File "%s" not deleted (because logging level is DEBUG)' % fh_out.name)
            #FIXME: test that rjson is installed earlier ?
            code = os.linesep.join([code_json_extract % {'magicvariable': magicvariable, 
                                                         'filename': fh_out.name},
                                    code])
            logger.debug(os.linesep.join(('Running R code snippet:',
                                          code)))
            returncode = self.run(code)
            if returncode != 0:
                raise Exception("R returned non-zero exit status %i" % returncode)
        return returncode

    def run(self, code,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE):
        """ R arbitrary code in a subprocess. 

        :param code: a string with R code

        Returns the return code for the child process.
        """
        cmd = (self.path, '--slave',)
        with open(os.devnull, "w") as fnull, \
             core.Popen(cmd, stdin=stdin,
                        stdout=stdout,
                        stderr=stderr) as proc:
            stdout, stderr = proc.communicate(input=code.encode('ascii'))
            if proc.returncode != 0:
                logger.warning(stderr)
            return proc.returncode
        

    def __repr__(self):
        s = [super(R, self).__repr__(), '  version: %s' % self.version]
        return '\n'.join(s)





# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import shutil, os, tempfile
from railroadtracks import environment
from railroadtracks.easy import qsub
from railroadtracks.easy import Project, rnaseq, Environment, TaskSet
from railroadtracks.model.simulate import PHAGEFASTA

SETUP = 'source %s' % os.path.abspath('setup.sh')
class QsubTestCase(unittest.TestCase):    
    def setUp(self):
        wd = tempfile.mkdtemp(dir=os.path.expanduser("~"))
        self.qsub_dir = tempfile.mkdtemp(dir=os.path.expanduser("~"))
        self.project = Project(rnaseq, wd=wd)

    def tearDown(self):
        # more complicated than it should: NFS and/or qsub seem to be leaving read-only locks behind
        wd = self.project.wd
        del(self.project)
        shutil.rmtree(wd)
        shutil.rmtree(self.qsub_dir)

    @unittest.skipIf(not environment.Executable.ispresent(qsub.QSUB_EXEC),
                     "No executable 'qsub' in the PATH.")
    def test_map(self):
        env = Environment(rnaseq)
        b2b = env.activities.INDEX.bowtie2build
        Assets = b2b.Assets
        assets = Assets(Assets.Source(rnaseq.FASTAFile(PHAGEFASTA)),
                        None)
        ts = TaskSet()
        task = self.project.add_task(b2b, assets)
        ts.add(task)
        task = self.project.add_task(b2b, assets, ('--ntoa',))
        ts.add(task)
        h2 = qsub.QsubExecution(SETUP, self.qsub_dir)
        h2.map(ts)
        # check that all tasks are complete
        for task in ts:
            self.assertEquals(hortator._TASK_DONE, task.status)


if __name__ == '__main__':
    unittest.main()

    #suite = unittest.TestLoader.loadTestsFromTestCase(QsubTestCase)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx
from collections import namedtuple, deque
from railroadtracks import easy
from railroadtracks.easy.execution import exec_taskfilter

TaskSetBundle = namedtuple('TaskSetBundle', 'label taskset mapper filter')

def get_tasksetid(taskset):
    """ Utility function that returns a valid (hashable) key for a task set.
    """
    return frozenset(x.task_id for x in taskset)

class TaskSetGraph(easy.TaskSet):
    """
    Graph of TaskSet nodes.
    """
    def __init__(self, project=None, defaultmapper=None, 
                 defaulttasksetfilter=exec_taskfilter):
        """
        :param project: 
        :param defaultmapper: default task mapper
        :param defaulttasksetfilter: default filter to apply to a taskset before passing it
                                     to the task mapper. This is not used for the moment.
        """
        super(TaskSetGraph, self).__init__(project=project)
        self._digraph = networkx.digraph.DiGraph()
        self._taskid2tasksetid = dict()
        self.defaultmapper = defaultmapper
        self.defaulttasksetfilter = defaulttasksetfilter

    @property
    def digraph(self):
        return self._digraph

    def _ensure_task_unique(self, taskset):
        intersect = self.intersection(taskset)
        if len(intersect) > 0:
            if len(intersect) > 1:
                plural = ('s', 'are')
            else:
                plural = ('', 'is')
            raise ValueError('%i of the task%s in the task set %s already in the graph.' % (len(intersect), plural[0], plural[1]))
    

    def add_taskset(self, taskset, label=None, mapper=None, filter=None):
        """ Add a TaskSet to the graph. 
        The dependency relationships with other :class:`TaskSet` objects
        are inferred from the assets in the individual :class:`Task` objects.

        :param taskset: a :class:`easy.tasks.TaskSet`
        :param label: a label for the `taskset`
        :param mapper: a task mapper to use with this taskset
        :param filter: a task filter to use this this taskset

        """
        self._ensure_taskset_project(taskset)
        self._ensure_task_unique(taskset)
        tasksetid = get_tasksetid(taskset)
        # make a copy to prevent problems if further changes in the TaskSet performed later
        # in the enclosing frame
        taskset_copy = easy.TaskSet()
        if mapper is None:
            mapper = self.defaultmapper
        self._digraph.add_node(tasksetid, attr_dict={'label': label,
                                                     'taskset': taskset_copy,
                                                     'mapper': mapper,
                                                     'filter': filter})
        for task in taskset:
            self._taskid2tasksetid[task.task_id] = tasksetid
            taskset_copy.add(task)
            super(TaskSetGraph, self).add(task)
            # parent tasks
            taskqueue = deque((x, tasksetid) for x in task.parent_tasks())
            while len(taskqueue) > 0:
                parent_task, child_tasksetid = taskqueue.pop()
                if parent_task.task_id in self._taskid2tasksetid:
                    # taskset in which the parent is present
                    p_tasksetid = self._taskid2tasksetid[parent_task.task_id]
                    if not self._digraph.has_edge(p_tasksetid, child_tasksetid):
                        self._digraph.add_edge(p_tasksetid, tasksetid)
                        # add the parents to the queue to reconstruct the graph of tasksets
                        for p in parent_task.parent_tasks():
                            taskqueue.appendleft((p, p_tasksetid))
                else:
                    pass
            # child tasks
            # (note: taskqueue is empty again when reaching here)
            taskqueue = deque((tasksetid, x) for x in task.child_tasks())
            while len(taskqueue) > 0:
                parent_tasksetid, child_task = taskqueue.pop()
                if child_task.task_id in self._taskid2tasksetid:
                    # taskset in which the child is present
                    c_tasksetid = self._taskid2tasksetid[child_task.task_id]
                    if not self._digraph.has_edge(tasksetid, c_tasksetid):
                        self._digraph.add_edge(tasksetid, c_tasksetid)
                        # add the children to the queue to reconstruct the graph of tasksets
                        for c in child_task.child_tasks():
                            taskqueue.appendleft((tasksetid, c))
                else:
                    pass
        return tasksetid

    def add(self, obj, label=None, mapper=None, filter=None):
        """ Generic add method, allowing the addition of either a Task or TaskSet object"""
        if isinstance(obj, easy.Task):
            obj = easy.TaskSet(iterable=(obj,))
        assert isinstance(obj, easy.TaskSet)
        self.add_taskset(obj, label=label, mapper=mapper)


    def execution_list(self):
        """ Return a list of :class:`TaskSetBundle` objects in an order in
        which they can be executed. """
        g = self._digraph
        return tuple(TaskSetBundle(g.node[x]['label'],
                                   g.node[x]['taskset'],
                                   g.node[x]['mapper'],
                                   g.node[x]['filter']) for x in networkx.topological_sort(g))

    def _filtered_taskset(self, tsb):
        if tsb.filter is None:
            taskset_filter = self.defaulttasksetfilter
        else:
            taskset_filter = tsb.taskset_filter
        if taskset_filter is None:
            taskset = tsb.taskset
        else:
            taskset = taskset_filter(tsb.taskset)
        return taskset
        
    def execute(self, 
                update_func=None):
        """ As with the method of the same name in the the parent class
        this will execute the tasks contained, with the notable difference
        that the execution that the tasks will be grouped according to the
        nested task sets defined and in the implicit order defined by the
        connections at the task level.

        This method is little beyond a wrapper around the method
        execution_list(), iterating through the task sets returned and
        executing the tasks with either the given task mapper of the default
        task mapper. 
        """
        lst = self.execution_list()
        for i, tsb in enumerate(lst):
            if tsb.mapper is None:
                mapper = self.defaultmapper
            else:
                mapper = tsb.mapper
            if mapper is None:
                raise ValueError('No task mapper was specified when the task was added, ' + \
                                 'and the default task mapper for the %s is not specified' % type(self).__name__)
            taskset = self._filtered_taskset(tsb)
            if update_func is not None:
                update_func(i, lst, taskset)
            mapper.map(taskset)

def update_func(i, lst, filtered_taskset):
    print('%i - %s (%i/%i tasks)' % (i, lst[i].label, len(filtered_taskset), len(lst[i].taskset)))



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple model for the expression analysis using RNA-Seq
"""
import os, sys, shutil
import re
import warnings
import csv
import zlib
import collections
import subprocess
import tempfile
import logging
logger = logging.getLogger(__name__)
import argparse
from . import environment
from railroadtracks.model.diffexp import RSOURCES_DIR
from . import core
import gzip
from railroadtracks.model import files
from railroadtracks.model.files import (File,
                                        FASTAFile,
                                        FASTQFile,
                                        GFFFile,
                                        FASTQPossiblyGzipCompressed, 
                                        BAMFile,
                                        SAMFile,
                                        BAMFileSortedByID,
                                        CSVFile,
                                        TSVFile,
                                        FilePattern,
                                        CSVFileSequence,
                                        SAMFileSequence,
                                        SamtoolsSamToBam,
                                        SamtoolsBamToSam,
                                        SamtoolsSorterByID,
                                        BEDFile,
                                        BedtoolsBamToFastq,
                                        BedtoolsBamToFastqPE)
from railroadtracks.model import simulate
from railroadtracks.model.simulate import (FluxsimulatorExpression,
                                           FluxsimulatorDerivedExpression,
                                           FluxsimulatorSequencing,
                                           FluxSimulatorParameters,
                                           FluxSimulatorParametersAndPro)
from railroadtracks.model import aligners
from railroadtracks.model.aligners import (BWAIndex,
                                           BWAIndexFiles,
                                           BWA,
                                           Bowtie,
                                           SavedBowtieIndex,
                                           Bowtie2,
                                           SavedBowtie2Index,
                                           BowtieBuild,
                                           Bowtie2Build,
                                           StarIndex,
                                           StarAlign,
                                           GsnapIndex,
                                           GsnapAlign,
                                           SailfishIndex,
                                           SalmonIndex,
                                           TopHat,
                                           TopHat2,
                                           SamtoolsExtractUnaligned)
from railroadtracks.model import diffexp
from railroadtracks.model.diffexp import (EdgeR,
                                          DESeq,
                                          DESeq2,
                                          LimmaVoom,
                                          EBSeq)

from railroadtracks.model import quantify
from railroadtracks.model.quantify import (SailfishQuant,
                                           SalmonAlignQuant,
                                           FeatureCount,
                                           HTSeqCount)

from railroadtracks.model import misc
from railroadtracks.model.misc import Anyscript

from railroadtracks.unifex import UnifexError, _cmdfromuei
from abc import ABCMeta, abstractproperty
            

class ACTIVITY(simulate.ACTIVITY, 
               aligners.ACTIVITY, 
               quantify.ACTIVITY, 
               diffexp.ACTIVITY,
               files.ACTIVITY,
               misc.ACTIVITY):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'SIMULATE INDEX ALIGN SORT QUANTIFY NORMALIZE DIFFEXP UTILITY MISC'
    NORMALIZE = 'Normalize'



class AssetsCRCHeadTail(core.AssetsStep):
    """
    Assets for :class:`CRCHeadTail`

    """
    Source = core.assetfactory('Source', [core.AssetAttr('file', File, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('crc', CSVFile, '')])

class CRCHeadTail(core.StepAbstract):
    """
    Compute a CRC32-based checksum on the beginning (head) and end (tail) of a file
    as a cheap way to check whether 2 files contain identical data.
    Might be useful with large files, however its main purpose is testing.
    """
    _name = 'crc-file'
    _default_execpath = None
    Assets = AssetsCRCHeadTail
    activities = (ACTIVITY.UTILITY, )

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        # no third-party executable involved
        if self._execpath is not None:
            raise ValueError('No third-party executable involved. The executable should be None')


    @property
    def version(self):
        from railroadtracks import __version__
        return __version__

    def run(self, assets, parameters = tuple()):
        # FIXME: should these checks be moved to the Assets class ?
        out_counts = assets.target.crc
        out_fh = open(out_counts.name, 'w')
        csv_w = csv.writer(out_fh)
        csv_w.writerow(['name', 'crc'])
        for name in assets.source.file:
            with open(name, 'rb') as in_fh:
                # begining of the file
                data = in_fh.read(10000)
                crc = zlib.crc32(data)
                # end of the file
                in_fh.seek(0, 2)
                n = in_fh.tell()
                if n < 10000:
                    in_fh.seek(0)
                else:
                    in_fh.seek(-10000, 2)
                data = in_fh.read(10000)
                crc = zlib.crc32(data, crc)                
            csv_w.writerow((name, crc))
        out_fh.flush()
        out_fh.close()
        cmd = None
        returncode = 0
        return (cmd, returncode)


# patch ngs_plumbing with class Gzip-friendly
try:
    import ngs_plumbing.fastq
    hasngsp = True
except:
    hasngsp = False
if hasngsp:
    import io
    #FIXME: This should ideally be fixed upstream in ngs_plumbing
    class GzipFastqFilePair(ngs_plumbing.fastq.FastqFilePair):
        """ Pair of FASTQ files
        The default iterator will go through the paired entries in the file
        (not the rows), assuming that they are in the same order.
        No check that this is the case (using IDs) is performed.
        """

        def __init__(self, filename1, filename2, **kwargs):
            Cls = type(self)
            if ngs_plumbing.fastq._python2:
                fh1 = gzip.open(filename=filename1, **kwargs)
                fh2 = gzip.open(filename=filename2, **kwargs)
            else:
                stream = io.FileIO(filename1)
                fh1 = gzip.open(stream, **kwargs)
                stream = io.FileIO(filename2)
                fh2 = gzip.open(stream, **kwargs)
            self._fh1 = fh1
            self._fh2 = fh2

        def __iter__(self):
            return ngs_plumbing.fastq.iter_entry_pairs(self._fh1, self._fh2)

        def iter_seqqual_pairs(self):
            """ Iterate over the pairs of (sequence+quality) in 
            the file. """
            return ngs_plumbing.fastq.iter_seqqual_pairs(self._fh1, self._fh2)

        def readcount(self):
            pos = self.tell()
            for i, entry in enumerate(self):
                pass
            return i+1
    




        

class AssetsColumnMerger(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('counts', CSVFileSequence, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('counts', CSVFile, '')])

class ColumnMerger(core.StepAbstract):
    Assets = AssetsColumnMerger
    _name = 'column-merge'
    _default_execpath = None
    activities = (ACTIVITY.UTILITY, )
    #activity = property(lambda x: x.activities, None, None)

    def __init__(self,executable=None):
        """ First parameter is the column used as ID. If None, no column used as ID.
        Second parameter is the column to merge."""
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable # no third-party executable involved
        if  self._execpath is not None:
            raise ValueError('The executable cannot be set. This is a fully internal step.')

    @property
    def version(self):
        from railroadtracks import __version__
        return __version__

    def run(self, assets, parameters = ('0','1')):
        # FIXME: should these checks be moved to the Assets class ?
        out_counts = assets.target.counts
        l = len(tuple(out_counts))
        if l > 1:
            raise ValueError('Merging columns can only have one target file.')
        else:
            import resource
            max_fh_soft, max_fh_hard = resource.getrlimit(resource.RLIMIT_NOFILE)
            if l > max_fh_hard:
                raise ValueError('On this system the maximum number of columns is'
                                 ' %i (and you are trying to merge %i columns)'
                                 ' [note: opened file handles should be taken out of the'
                                  ' maximum number]' % (max_fh_hard-1, l))
        in_counts = assets.source.counts
        # open all
        it_in_counts = in_counts.iteritems()
        cls, value = next(it_in_counts)
        head = cls(value)
        head_iter = head.iter_entry()
        tail = tuple(cls(value) for cls, value in it_in_counts)
        tail_iter = tuple(x.iter_entry() for x in tail)
        # merge
        if len(parameters) != 2:
            raise ValueError('There should be 2 parameters: the index of the columns with IDs, the index of the column with values.')
        try:
            idcol_i = int(parameters[0])
        except ValueError:
            if parameters[0] == 'None':
                idcol_i = None
            else:
                ValueError('The first parameter should be string representation of an integer or "None".')
        valcol_i = int(parameters[1])
        with open(out_counts.name, 'w') as out_fh:
            csv_w = csv.writer(out_fh)                    
            if idcol_i is None:
                for row in head_iter:
                    counts = [row[valcol_i]]
                    for col_i, h in enumerate(tail_iter):
                        row = next(h)
                        counts.append(row[valcol_i].rstrip())
                    csv_w.writerow(counts)
            else:
                for row in head_iter:
                    identifier_head = row[idcol_i]
                    counts = [row[valcol_i].rstrip()]
                    for col_i, h in enumerate(tail_iter):
                        row = next(h)
                        identifier = row[idcol_i]
                        if identifier == identifier_head:
                            counts.append(row[valcol_i].rstrip())
                        else:
                            raise ValueError('Mismatching IDs (first columns has "%s"'
                                             ' while column %i has "%s")' % (identifier_head,
                                                                             col_i+1, identifier))
                    csv_w.writerow([identifier_head, ] + counts)
        
        cmd = None
        returncode = 0
        return (cmd, returncode)

class AssetsNormalizer(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('count', CSVFile, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('normalizedcounts', CSVFile, '')])

class Normalizer(core.StepAbstract):
    Assets = AssetsNormalizer


def _picard_getversion(execpath):
    cmd = [execpath, '--version']
    try:
        logger.debug(subprocess.list2cmdline(cmd))
        res = subprocess.check_output(cmd)
    except OSError as ose:
        raise UnifexError("""Command: %s
        %s""" % (' '.join(cmd), ose))

    m = re.match('^([0-9].+)$', res)
    if m is None:
        raise RuntimeError('Could not find the version number.')
    return m.groups()[0]



class PicardCollectAlignmentSummaryMetrics(core.StepAbstract):
    activities = (ACTIVITY.UTILITY, )
    _name = 'picard-collect-alignment-summary-metrics'
    _default_execpath = 'picard-collect-alignment-summary-metrics.sh'

    class Assets(core.AssetsStep):
        Source = core.assetfactory('Source', [core.AssetAttr('alignedreads', BAMFileSortedByID, ''),
                                              core.AssetAttr('reference', FASTAFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('metrics', FilePattern, '')])

    parser = argparse.ArgumentParser(_name)
    parser.add_argument('ASSUME_SORTED',
                        default = 'false',
                        choices = ('true', 'false'))

    def __init__(self, execpath=None):
        """        
        :param executable: path to the executable
        :type param: :class:`str`
        """
        if execpath is None:
            execpath = type(self)._default_execpath
        self._execpath = execpath
        self._version = None
    
    @property 
    def version(self):
        if self._version is None:
            self._version = _picard_getversion(self.execpath)
        return self._version

    def run(self, assets, parameters = tuple()):
        options, unknown = self.parser.parse_known_args()
        # FIXME: check that input is sorted BAM. If not, forbid the use of ASSUME_SORTED ?

        # build command line
        cmd = [self._execpath, ]
        cmd.extend(parameters)
        cmd.extend(['INPUT_FILE=%s', assets.source.annotationfile.name,
                    'REFERENCE_SEQUENCE=%s', assets.source.reference.name,
                    'OUTPUT=%s', assets.target.metrics.name])

        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull:
            returncode = subprocess.check_call(cmd)

        return (cmd, returncode)




# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import os
import abc
from operator import attrgetter
import sys
import subprocess

from six import with_metaclass

if sys.version_info[0] < 3:
    class Popen(subprocess.Popen):
        def __enter__(self):
            return self
        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.stdout:
                self.stdout.close()
            if self.stderr:
                self.stderr.close()
            if self.stdin:
                self.stdin.close()
            self.wait()
else:
    # Python 3 !
    unicode = str
    Popen = subprocess.Popen

from flufl.enum import Enum


# Error message 
_NOTIMPLEMENTED_ABSTRACT = "This is an interface, and should be implemented by child classses."


class SavedEntityAbstract(with_metaclass(abc.ABCMeta, object)):
    """
    Represent a file, or group of files, whether already existing on the disk or not.

    Instances of the class are iterable in order to have a unified handling for single
    files and sequences of files.
    """

    _defined = False
    _name = False
    _extension = None

    def __init__(self, *args, **kwargs):
        raise TypeError('The class %s does not have a constructor.' % str(type(self)))

    def isnewerthan(self, savedentityabstract):
        """ """

        assert self._defined

        t_self = self.lastmodified()
        t = savedentityabstract.lastmodified()
        # special cases with None
        if t_self is None:
            # This is a problematic situation
            # savedentityabstract cannot be newer
            # than "self", that is still be materialized
            return False
        elif t is None:
            # t_self cannot be None (already tested above)
            # self is the newest then
            return True
        return t_self > t

    @staticmethod
    def _hash_components(cls, name):
        clsname = cls.__module__ + '.' + cls.__name__
        return (clsname, name)

    @property
    def hashdb(self):
        if not self._defined:
            raise ValueError("The hashdb for an undefined saved entity cannot be computed.")
        return hash(type(self)._hash_components(type(self), self._name))

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError(_NOTIMPLEMENTED_ABSTRACT)

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError(_NOTIMPLEMENTED_ABSTRACT)

    # FIXME: does it make any sense to keep this method at the top abstract level,
    #        or should it be cleaner to only have this for containers (sequences) ?
    @abc.abstractmethod
    def iteritems(self):
        raise NotImplementedError(_NOTIMPLEMENTED_ABSTRACT)


def _check_extension(name, extension):
    if (name is not None) and (extension is not None):
        ok = any(name.endswith(x) for x in extension)
        assert ok, \
            "The file name %s should have one of the following extensions: %s" % (name, repr(extension))

class File(SavedEntityAbstract):
    """
    File name, or prospective file name.

    This differs from regular files because it may represent a file that is not yet existing on the disk.
    The reason for this is to able to model the case where the input, or output, from a step is not yet
    existing.

    The class implements __iter__ to allow duck typing with instances of this class and with
    instances of the sibling class FilePattern.
    """

    def __init__(self,
                 name=None):
        """
        :param name: a string, or a sequence of strings with only one string
        """
        if name is not None:
            self._defined = True
            if not (isinstance(name, str) or isinstance(name, unicode)):
                if len(name) == 1:
                    name = next(iter(name))
                else:
                    raise ValueError('"name" should be a string.')
            _check_extension(name, self._extension)
        self._name = name

    def _getname(self):
        assert self._defined
        return self._name

    def _setname(self, name):
        assert not self._defined
        _check_extension(name, self._extension)
        self._name = name
        self._defined = True

    name = property(_getname, _setname, None,
                    'Path to the file. When not defined, this can be None and accessing it will raise an AssertionError (it can only be set then).')

    def lastmodified(self):
        assert self._defined
        if os.path.exists(self.name):
            return os.path.getmtime(self.name)
        else:
            return None

    def __len__(self):
        return 1

    def __iter__(self):
        """ Iterate through _the_ filename. """
        assert self._defined
        yield self.name

    def iteritems(self):
        """ """
        assert self._defined
        yield(type(self), self.name)

class FileSequenceAbstract(SavedEntityAbstract):
    """ Sequence of files """
    _type = SavedEntityAbstract

    @property
    def hashdb(self):
        return hash((super(FileSequenceAbstract, self).hashdb, 
                     tuple(x for x in self)))

    def __init__(self, savedentities):
        savesentities = tuple(savedentities)
        if savedentities is not None:
            # sanity checks
            savedentities = tuple(savedentities)
            for x in savedentities:
                # FIXME: check that there is the _same_ type for all items ? 
                assert isinstance(x, self._type), 'Expected type "%s" but got "%s"' % (str(self._type), str(type(x)))
                assert not isinstance(x, type(self)) # no recursive FileSequence allowed
            self._defined = True
        else:
            pass
            #FIXME: do we allow undefined FileSequences ???
        self._savedentities = savedentities

    def __len__(self):
        return len(self._savedentities)

    def __iter__(self):
        """ Iterate through files in the pattern """
        assert self._defined
        return iter(x.name for x in self._savedentities)

    def iteritems(self):
        assert self._defined
        for se in self._savedentities:
            yield (type(se), se.name)



class FileSequence(FileSequenceAbstract):
    _type = File

#FIXME: why isn't this in unifex.py ? circular reference ?
UnifiedExecInfo = collections.namedtuple('UnifiedExecInfo', 
                                         'executable model source target parameters logging_file logging_level')


# -- stepabstract-begin
class StepAbstract(with_metaclass(abc.ABCMeta, object)):
    """ Abstract parent for steps. """

    # monicker under which the step will be known (must be unique within a step list)
    _name = abc.abstractproperty()
    # default name for executable associated with the class
    _default_execpath = abc.abstractproperty()
    
    # class of assets for the step (must be a child of :class:`AssetsStep`)
    Assets = abc.abstractproperty()

    activities = abc.abstractproperty(None, None, None, 
                                      "Activities associated with the step.")

    version = abc.abstractproperty(None, None, None,
                                   "Version of the executable associated with the step.")


    @property
    def hashdb(self):
        return hash((type(self), self._default_execpath, self.version, self.activities))

    @abc.abstractmethod
    def run(self, assets, parameters=tuple()):
        """ 
        :param assets: Assets to run the step with
        :type assets: :class:`AssetsStep`
        :param parameters: optional parameters
        """
        raise NotImplementedError(_NOTIMPLEMENTED_ABSTRACT)

    def uei(self, assets, parameters = tuple()):
        # FIXME: should return the unified execution command line
        uei = UnifiedExecInfo(self._execpath, self._name,
                              assets.source, assets.target, parameters,
                              None, None) #logging_file and logging_level
        return uei
# -- stepabstract-end

# iterate over the specification of sources
# name: name of the attribute
# cls: expected class for the attribute
# pattern: regular expression
class AssetAttr(object):
    """ Describe what an asset (as a attribute in an :class:`core.AssetsStep`) should be like. """
    __slots__ = ('name', 'cls', 'pattern', 'allownone')
    def __init__(self, name, cls, pattern, allownone=False):
        self.name = name
        self.cls = cls
        self.pattern = pattern
        self.allownone = allownone


#FIXME: _sources is not a good variable name (because an AssetSet is either called Source or Target) 
class AssetMetaReserved(Enum):
    """ Reserved attribute names for AssetMeta type
    (made as a central reference to facilitate refactoring in the choice
    of names proves unlucky later on)"""
    __order__ = 'SOURCES FIELDS'
    SOURCES = '_sources'
    FIELDS = '_fields' # don't touch this one in any case - it is providing compatibility with namedtuples

def _pset_factory(name):
    def func(self, value):
        if getattr(self, name) is not None:
            raise AttributeError("'%s' object attribute '%s' can only be set if uninitialized." % (type(self), name))
        privatename = '_' + name
        setattr(self, privatename, value)

class AssetMeta(type):
    """ Meta-class looking for a property "AssetMetaReserved.SOURCES",
    expected to be a sequence of objects with the attributes

    :param name: name for the asset
    :type name: class:`str`
    :param cls: class inheriting from SavedEntityAbstract
    :type cls: type
    :param pattern: pattern for file name
    :type pattern: class:`str`
    """
    def __new__(cls, name, parents, dct):
        _sources = dct.get(AssetMetaReserved.SOURCES.value)
        if _sources is None:
            _sources = collections.OrderedDict()
            dct[AssetMetaReserved.SOURCES.value] = _sources
        
        reserved = set(AssetMetaReserved.__dict__.keys())
        _fields = list()
        for aa in _sources:
            assert isinstance(aa, AssetAttr)
            privatename = '_'+aa.name
            # sanity check
            if (aa.name in reserved) or (privatename in reserved):
                raise ValueError('The attribute name "%s" is reserved and cannot be used.' % aa.name)
            elif (aa.name in dct) or (privatename in dct):
                # duplicated name
                raise ValueError("The name %s is defined more than once" % aa.name)
            assert isinstance(aa.name, str)
            assert issubclass(aa.cls, SavedEntityAbstract)
            assert isinstance(aa.pattern, str)
            # add to the class for tab-completion (whenever a user is looking for hints),
            # # but do not allow the wrong type to be set
            dct[privatename] =  None
            dct[aa.name] = property(attrgetter(privatename),
                                    None, None)
            # add to _fields (so it is looking like a named tuple)
            _fields.append(aa.name)
        #
        dct[AssetMetaReserved.FIELDS.value] = _fields
        # complete the initialization
        return super(AssetMeta, cls).__new__(cls, name, parents, dct)

    def __repr__(cls):
        res = [super(AssetMeta, cls).__repr__(),
               'The fields for the set of assets are:']
        for f, v in zip(cls._fields, getattr(cls, AssetMetaReserved.SOURCES.value)):
            res.append('    - %s (%s)' % (f, v.cls.__name__))
        return os.linesep.join(res)

#FIXME: implement a deferred check (to allow unspecified asset elements to exist)
class AssetSet(with_metaclass(AssetMeta, object)):
    """ Ordered set of assets """

    @property
    def hashdb(self):
        lst = list()
        for x in self._sources:
            val = getattr(self, x.name)
            if val is None:
                if not x.allownone:
                    raise ValueError("Invalid asset")
                hashsb = val
            else:
                hashdb = val.hashdb
            lst.append((x.name, hashdb))
        return hash(tuple(lst))

    def __init__(self, *args):
        sources = getattr(self, AssetMetaReserved.SOURCES.value)
        assert len(args) == len(sources), 'The following parameter(s) must be specified, and in that order: %s' % \
            str(tuple(x.name for x in sources))
        for s,value in zip(sources, args):
            if (value is None) and (s.allownone):
                 continue
            assert isinstance(value, s.cls), 'The object "%s" was expected to be of type %s (but is a %s)' % \
                (repr(value), repr(s.cls), type(value))
            privatename = '_'+s.name
            setattr(self, privatename, value)
        pass

    def __str__(self):
        l = len(self)
        res = [super(AssetSet, self).__repr__(),
               '  with %i element(s) (-: defined, *: undefined):' % l]
        for i in range(min(5,l)):
            if getattr(self, self._sources[i].name)._defined:
                bullet = '-'
            else:
                bullet = '*'
            res.append('  %s `%s` (a %s)' % (bullet, self._sources[i].name, self._sources[i].cls))
        if l > 5:
            res.append('    ...')
        return os.linesep.join(res)
        
    def __len__(self):
        """
        Number of assets in the set 
        :rtype: :class:`int`
        """
        return len(self._sources)

    def __iter__(self):
        for s in self._sources:
            yield getattr(self, s.name)

    @classmethod
    def createundefined(cls):
        """
        Create an unspecified AssetsSet-inheriting instance.
        This is an helper function for the situations where
        the AssetSet cannot be fully specified (for example, it is
        not yet known what will the target files be.
        
        :param obj: object such as an instance of Source or Target
        :param classname: class name for the tuple
        """
        # loop over AssetAttr elements
        args = list()
        for aa in getattr(cls, AssetMetaReserved.SOURCES.value):
            # aa.name
            # aa.cls
            # aa.pattern
            args.append(aa.cls(None))
        res = cls(*args)
        return res

    @classmethod
    def getassetattr(cls, attrname):
        """ Get the 'AssetAttr' for an attribute name. """
        try:
            i = cls._fields.index(attrname)
        except ValueError as ve:
            raise ValueError('The attribute "%s" is not in "%s".' % (attrname, cls.__name__))
        return cls._sources[i]

def assetfactory(name, savedentities):
    cls = AssetMeta(name, (AssetSet, ), {AssetMetaReserved.SOURCES.value: savedentities})
    return cls

class AssetsStep(object):
    """
    Assets used by any step are split into sources,
    and targets.
    """
    Source = assetfactory('Source', [])
    Target = assetfactory('Target', [])

    @property
    def hashdb(self):
        return hash((self.source.hashdb, self.target.hashdb))

    def __init__(self, source, target=None):
        """
        :param source: 
        :type source: :class:`AssetSet`
        :param target:
        :type target: :class:`AssetSet`
        """
        # sanity check before storing into the instance
        self.source = self._check(source, self.Source)
        if target is None:
            # allow unspecified targets to exists.
            # build the asset as undefined
            self.target = self.Target.createundefined()
        else:
            self.target = self._check(target, self.Target)

    def _check(self, obj, cls):
        """
        :param obj: object such as an instance of Source or Target
        :param cls: class inheriting from :class:`AssetSet`
        """
        # try:
        #     getattr(cls, AssetMetaReserved.SOURCES)
        # except:
        #     import pdb; pdb.set_trace()
        for assetattr, field in zip(getattr(cls, AssetMetaReserved.SOURCES.value), cls._fields):
            if assetattr.allownone:
                continue
            assert hasattr(obj, field), '%s is missing "%s"' % (str(cls), field)
            field_attr = getattr(obj, field)
            assert isinstance(field_attr,
                              SavedEntityAbstract), \
                '%s is not an instance of %s for class %s (got %s)' % \
                (field, SavedEntityAbstract,
                 type(self), type(field_attr))
        return obj

    def __str__(self):
        return os.linesep.join((super(AssetsStep, self).__str__(), 
                                str(self.source), 
                                '---', 
                                str(self.target)))


def steplist(model):
    """ Return the list of steps (as a list of classes inheriting from core.StepAbstract) 
    present in a a model (a namespace such as a module).
    
    If the name _STEPLIST_CLASSES (a list of classe) is present in the model, it will be 
    used as a definition of the exported classes. If not the content of the namespace will
    be inspected and any class inheriting from StepAbstract will be returned.

    """
    if hasattr(model, '_STEPLIST_CLASSES'):
        classlist = model._STEPLIST_CLASSES
    else:
        # catch all steps
        import inspect
        def is_Step(x):
            return inspect.isclass(x) \
                and (not inspect.isabstract(x)) \
                and issubclass(x, (StepAbstract,))

        classlist = inspect.getmembers(model, 
                                       is_Step)
        classlist = ([x[1] for x in classlist])
    return classlist

class DEFAULT_ACTIVITY(Enum):
    MISC = 'Misc.'

class Model(object):
    """ Namespace for a model. """
    def __init__(self, 
                 steplist_classes = tuple(),
                 activity = DEFAULT_ACTIVITY):
        self._STEPLIST_CLASSES = tuple(steplist_classes)
        self.ACTIVITY = activity
        



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from railroadtracks import hortator
from railroadtracks.easy.tasks import Task
from collections import namedtuple, deque
import networkx

Node = namedtuple('Node', 'clsname id')


def _task(id, project, cursor):
    return project.get_task(id)

def _asset(id, project, cursor, assettype):
    """ Create a StoredEntity from its ID. """
    sql = """
    SELECT id, label, classname, entityname
    FROM stored_entity
    JOIN step_concrete2%sfile
    ON stored_entity.id=stored_entity_id
    WHERE stored_entity_id=?
    """ % assettype
    cursor.execute(sql, (id, ))
    dbid, label, clsname, entityname = cursor.fetchone()        
    return hortator.StoredEntity(dbid, label, clsname, entityname)

def _assetsequence(id, project, cursor, assettype):
    """ Create a StoredSequence from its ID. """
    sql = """
    SELECT stored_sequence.id, label, stored_sequence.classname, se2s.stored_entity_id
    FROM stored_sequence
    JOIN stored_entity2sequence as se2s
    ON se2s.stored_sequence_id=stored_sequence.id
    JOIN step_concrete2%sfile as sc2sf
    ON sc2sf.stored_sequence_id=se2s.stored_sequence_id
    WHERE se2s.stored_sequence_id=?
    ORDER BY pos
    """ % assettype
    cursor.execute(sql, (id, ))
    storedentities = list()
    for dbid, label, clsname, assetid in cursor:
        storedentities.append(assetid)
            
    return hortator.StoredSequence(dbid, label, clsname, storedentities)

nodeclsname2constructor = {'Task': _task,
                           'Asset': _asset,
                           'AssetSequence': _assetsequence}

def get_digraph(project):
    digraph = networkx.digraph.DiGraph()
    #
    cursor = project.persistent_graph.connection.cursor()
    sql_tasks_template = """
    SELECT step_concrete.id,
           stored_entity_id,
           stored_sequence_id
    FROM step_concrete,
         step_concrete2%sfile
    WHERE step_concrete2%sfile.step_concrete_id=step_concrete.id
    """
    for assettype in ('src', 'target'):
        sql_tasks = sql_tasks_template % (assettype, assettype)
        cursor.execute(sql_tasks)
        #
        cursor_nodes = project.persistent_graph.connection.cursor()
        for taskid, assetid, assetsequenceid in cursor:
            nodeclsname = 'Task'
            task_node = Node(nodeclsname, taskid)
            if task_node not in digraph:
                instance = nodeclsname2constructor[nodeclsname](taskid, project, cursor_nodes)
                digraph.add_node(task_node, instance=instance)
            # sanity check
            assert (assetid is None) != (assetsequenceid is None), \
                                        'Issue with the database: it is either an asset or a sequence of assets.'
            if assetsequenceid is None:
                nodeclsname = 'Asset'
                dbid = assetid
                asset_node = Node(nodeclsname, dbid)
            else:
                # no need to test because of the sanity check above
                nodeclsname = 'AssetSequence'
                dbid = assetsequenceid
                asset_node = Node(nodeclsname, dbid)
            if asset_node not in digraph:
                instance = nodeclsname2constructor[nodeclsname](dbid, project, cursor_nodes, assettype)
                digraph.add_node(asset_node, instance=instance)
            if assettype == 'src':
                digraph.add_edge(asset_node, task_node)
            else:
                digraph.add_edge(task_node, asset_node)
    #
    sql_seq = """
    SELECT se2s.stored_sequence_id as assetsequenceid,
           stored_entity.id as assetid,
           pos
    FROM stored_entity2sequence AS se2s
    JOIN stored_entity
    ON se2s.stored_entity_id=stored_entity.id
    ORDER BY assetsequenceid, pos
    """
    cursor.execute(sql_seq)
    for assetsequenceid, assetid, pos in cursor:
        assetsequence_node = Node('AssetSequence', assetsequenceid)
        asset_node = Node('Asset', assetid)
        if asset_node not in digraph:
            instance = nodeclsname2constructor[nodeclsname](assetid, project, cursor_nodes, 'Asset')
            digraph.add_node(asset_node, instance=instance)
        digraph.add_edge(asset_node, assetsequence_node)
    #

    return digraph

class TaskGraph(object):
    
    def __init__(self, project):
        self._project = project
        self._digraph = get_digraph(project)

    @property
    def digraph(self):
        return self._digraph

    def predecessor_tasks(self, task):
        if isinstance(task, Task):
            node = Node('Task', task.task_id)
        else:
            node = task
        digraph = self._digraph
        assets = digraph.predecessors(node)
        ptasks = list()
        while len(assets) > 0:
            a = assets.pop()
            if a.clsname == 'AssetSequence':
                assets.extend(digraph.predecessors(Node('AssetSequence', a.id)))
                continue
            for p in digraph.predecessors(a):
                assert p.clsname == 'Task'
                ptasks.append(p)
        return ptasks



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import logging
logger = logging.getLogger(__name__)

from railroadtracks import core

from railroadtracks.model.files import FilePattern

class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'MISC'
    MISC = 'Miscellaneous'

class AssetsAnyscript(core.AssetsStep):
    """
    Assets for :class:`Anyscript`

    """
    Source = core.assetfactory('Source', [core.AssetAttr('script', core.File, ''),
                                          core.AssetAttr('input', core.FileSequence, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('output', FilePattern, '')])

class Anyscript(core.StepAbstract):
    """
    Do anything.
    """
    _name = 'anyscript'
    _default_execpath = None
    Assets = AssetsAnyscript
    activities = (ACTIVITY.MISC, )

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        # no third-party executable involved
        if self._execpath is not None:
            raise ValueError('No third-party executable involved. The executable cannot be None')

    @property
    def version(self):
        from railroadtracks import __version__
        return __version__

    def run(self, assets, parameters = tuple()):
        executable = assets.source.script.name
        # FIXME: check that this is an executable
        assert os.path.exists(executable)
        assert os.path.exists(assets.target.output.name)
        outdir = assets.target.output.name
        assert os.path.isdir(outdir)
        curdir = os.getcwd()
        os.chdir(outdir)
        try:
            cmd = [executable, ]
            cmd.extend(x.name for x in assets.source.input)
            with open(os.devnull, 'w') as fnull:
                subprocess.check_call(cmd, sdtout = fnull, stderr = fnull)
        finally:
            os.chdir(curdir)
        return (cmd, 0)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import tempfile, shutil, gzip

from railroadtracks import core
import railroadtracks.model.simulate



class AssetsTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_assetfactory(self):
        Foo = core.assetfactory('Foo', [core.AssetAttr('bar', core.File, '')])
        # check that the asset bar is defined
        foo = Foo(core.File(''))
        # check that an incorrect type raises an error
        self.assertRaises(AssertionError, Foo, 123)
        self.assertTrue(foo.bar._defined)
        # check that incorrect parameters make it fail
        self.assertRaises(AssertionError,
                          core.assetfactory, 
                          'Foo', [core.AssetAttr('bar', core.File, None)])
        # check that trying to modify an asset raises an error
        self.assertRaises(AttributeError, setattr, foo, 'bar', 123)


    def test_AssetStep_NoTarget(self):
        Foo = core.assetfactory('Foo', [core.AssetAttr('bar', core.File, '')])
        # no target
        assetset = core.AssetsStep(Foo)
        self.assertTrue(isinstance(assetset.__str__(), str))

    def test_AssetStep_Target(self):
        Foo = core.assetfactory('Foo', [core.AssetAttr('bar', core.File, '')])
        Bar = core.assetfactory('Bar', [core.AssetAttr('foo', core.File, '')])
        assetset = core.AssetsStep(Foo, Bar)
        self.assertTrue(isinstance(assetset.__str__(), str))

    def test_assetfactory_allownone(self):
        # check that allownone=True allows unspecified assets
        Foo = core.assetfactory('Foo', [core.AssetAttr('bar', core.File, '', allownone=True)])
        # check that the asset bar is defined
        foo = Foo(core.File(''))
        self.assertTrue(foo.bar._defined)
        # check that an incorrect type raises an error
        self.assertRaises(AssertionError, Foo, 123)
        foo = Foo(None)
        self.assertTrue(foo.bar is None)
        # check that trying to modify an asset raises an error
        self.assertRaises(AttributeError, setattr, foo, 'bar', 123)

    def test_createundefined(self):
        # -- createundefined-begin
        Foo = core.assetfactory('Foo', [core.AssetAttr('bar', core.File, '')])
        # create an undefined set of assets of type Foo
        undefoo = Foo.createundefined()
        # -- createundefined-end
        self.assertTrue(isinstance(undefoo, Foo))
        # check that the asset bar is an undefined "saved entity"
        self.assertFalse(undefoo.bar._defined)
        # check that trying to modify an asset raises an error
        # (modifying an asset value will be possible though)
        self.assertRaises(AttributeError, setattr, undefoo, 'bar', 123)
        undefoo.bar.name = '123'
        self.assertEqual('123', undefoo.bar.name)


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Persistence/memoization for the DAG
"""

from collections import namedtuple, Counter
from importlib import import_module
from railroadtracks.core import StepAbstract
from railroadtracks import __version__
import collections
import json
import os, shutil
import subprocess
import uuid
import warnings
import itertools
import operator
import networkx
import time
from . import core
from . import unifex

import enum
if hasattr(enum, '__version__') and enum.__version__.startswith('0.4'):
    raise ImportError("""
The Python package 'enum34' is required. Unfortunately, whenever the package 'enum' is also
present it can mask 'enum34'. Make sure that both 'enum34' is installed and 'enum'
is uninstalled.
    """)
Enum = enum.Enum

#FIXME: Shouldn't the definition come from separate sources (so the coupling between creation and access
# to the persistance layer is not too much tied to this Python script ?
_TASK_TODO = 'To do'
_TASK_DONE = 'Done'
_TASK_INPROGRESS = 'Work-in-progress'
_TASK_FAILED = 'Failed'
_TASK_STATUS_LIST = {_TASK_TODO: None, 
                     _TASK_INPROGRESS: None,
                     _TASK_DONE: None,
                     _TASK_FAILED: None}

# structure to have database ID, and flag to indicate if new (just created)
DbID = namedtuple('DbID', 'id new')

class Step:
    """
    When used in the context of a StepGraph, a Step is small graph,
    or subgraph, consituted of a vertex, connected downstream to targets and upstream
    to sources. For more information about a StepGraph, see
    the documentation for it.
    """
    __slots__ = ['step', 'sources', 'targets', 'parameters', 'model']
    def __init__(self, step, 
                 sources, targets, parameters,
                 model):
        """
        :param step: A concrete step in the database
        :type step: StepConcrete_DbEntry
        :param sources: a sequence of source assets
        :param targets: a sequence of target assets
        :param model: a model (e.g., :mod:`railroadtracks.rnaseq`)
        """
        assert isinstance(step, StepConcrete_DbEntry)
        self.step = step
        self.sources = sources
        self.targets = targets
        self.parameters = parameters
        self.model = model

    clsname = property(lambda self: self.step.clsname, None, None)
    unifiedname = property(lambda self: getattr(self.model, self.clsname)._name,
                           None, None)
    def iscomplete(self):
        return self.step.status == _TASK_DONE

    def run(self):
        # FIXME: check matching version numbers
        # self.step.version

        sources = ['='.join((x.label, x.entityname)) for x in self.sources]
        targets = ['='.join((x.label, x.entityname)) for x in self.targets]

        uei = core.UnifiedExecInfo(self.step.executable, self.unifiedname,
                                   sources,
                                   targets,
                                   self.parameters,
                                   None, None # logging_file, logging_level
                               )

        res = unifex.unified_exec_run(uei, unifex._make_stepdict(self.model))
        return res



class GRAPHDISPLAY(Enum):
    STEPLEVEL = {'layout': 'dot',
                 'layoutargs': '',
                 'kwargs': {'se_nodeid': '%(clsname)s_%(label)s',
                            'se_nodelabel': '{%(clsname)s | [%(label)s]}',
                            'step_nodeid': '%(clsname)s',
                            'step_nodelabel': '%(clsname)s'}}
    TASKLEVEL = {'layout': 'dot',
                 'layoutargs': '-Goverlap=prism -Gratio=1',
                 'kwargs': {'se_nodeid': '%(id)i | %(clsname)s',
                            'se_nodelabel': '%(id)i | {%(clsname)s|[%(label)s]}',
                            'step_nodeid': '%(id)i | %(clsname)s',
                            'step_nodelabel': '%(id)i | %(clsname)s'}}

# The DAG will be build from 2 types of vertices:
# - steps
# - assets
# Edges will always link of links of 2 different types
# FIXME: if the underlying database is changing, this is not updated
# (have a lock ? monitor changes ?)
class StepGraph(object):
    """ 
    The steps to be performed are stored in a directed acyclic graph (DAG).

    This graph can be thought of as a two-level graph. The higher level represents
    the connectivity between steps (we will call supersteps), and the lower-level expands each step
    into sources, targets, and a step using sources to produce targets.
    
    There is a persistent representation (currently a mysql database),
    and this class is aiming at isolating this implementation detail from a user.
    """

    def __init__(self, persistent_graph):
        # DAG used to resolve order in which steps should be run
        dag = networkx.DiGraph()
        self._dag = dag
        self._persistent_graph = persistent_graph
        # cache of DB IDs
        self._cache_dbid = dict()
        self._cache_stepvariant_dbid = dict()
        # build graph from the persistence layer
        for step in self._persistent_graph.iter_steps():
            # step contrains StepConcrete_DbEntry, sources, targets, parameters)
            self._update_graph(step.step.step_variant_id, step.step.id, step.sources, step.targets)

    @staticmethod
    def stepconcrete_dirname(stepconcrete_id):
        """
        Name of the directory corresponding to an ID.

        :param stepconcrete_id: ID for a directory.
        """
        stepconcrete_dirname = 'step_%i' % stepconcrete_id.id
        return stepconcrete_dirname

    def add(self, step, assets, parameters=tuple(), tag=1, use_cache=True):
        """ Add a step, associated assets, and optional parameters, to the StepGraph.

        The task graph is like a directed (presumably) acyclic multilevel graph.
        Asset vertices are only connected to step vertices (in other words asset vertices
        represent connective layers between steps).

        :param step: The step to be added
        :type step: a :class:`core.StepAbstract` (or of child classes) object
        :param assets: The assets linked to the step added. If :attr:`assets.target`
                       is undefined, the method will define it with unique identifiers
                       and these will assigned in place.
        :type assets: a :class:`core.AssetStep` (or of child classes) object
        :param parameters: Parameters for the step
        :type parameters: A sequence of :class:`str` elements
        :param tag: A tag to differentiate repetitions of the exact same task.
        
        :rtype: :class:`StepConcrete_DbEntry` as the entry added to the database
        """

        # add assets (note: the _cache ensures uniqueness)
        # add step (note: the _cache ensures uniqueness)
        assert isinstance(step, StepAbstract)

        #
        if use_cache:
            task_hashdb = (step.hashdb, assets.source.hashdb, parameters, tag)
            dbid = self._cache_dbid.get(task_hashdb)
            if dbid is not None:
                # We are not done yet as we need to recover the asset.target values
                for known_target in self._persistent_graph.get_targetassets(dbid):
                    candidate_target = getattr(assets.target, known_target.label)
                    if candidate_target._defined:
                        assert candidate_target.name == known_target.name, "Mismatch between target assets for '%s': previously '%s', now '%s'" % (known_target.label, known_target.entityname, candidate_target.name)
                    else:
                        # update the definition of the target
                        candidate_target.name = known_target.entityname
                    
                return DbID(dbid, False)
            

        # obtain the id for variant
        dbid = None
        if use_cache:
            stepvariant_hashdb = (step.hashdb, tuple(x.value for x in step.activities))
            dbid = self._cache_stepvariant_dbid.get(stepvariant_hashdb)
        if dbid is None:
            id_stepvariant = self._persistent_graph.id_step_variant(step,
                                                                    step.activities).id

        # undefined sources is not accepted
        # (exceptions being the source assets that are optionally None)
        if any((not y.allownone and not x._defined) for x,y in zip(assets.source, assets.source._sources)):
            raise ValueError('All sources must be defined.')

        # retrieve or create the task
        stepconcrete_id = self._persistent_graph.id_stepconcrete(id_stepvariant,
                                                                 assets.source,
                                                                 assets.target,
                                                                 parameters,
                                                                 tag = tag)

        stepconcrete_dirname = self.stepconcrete_dirname(stepconcrete_id)
        absdirname = os.path.join(self._persistent_graph._wd, stepconcrete_dirname)
        if stepconcrete_id.new:
            if os.path.isdir(absdirname):
                raise IOError("The directory %s is already existing." % absdirname)
            else:
                os.mkdir(absdirname)
        else:
            if not os.path.isdir(absdirname):
                raise IOError("The directory %s is missing." % absdirname)

        # loop over targets
        #    # targets_db = list()
        #    # for (label, asset) in zip(targets._fields, targets):
        #    #     # each asset can represent several saved objects
        #    #     entity_ids = tuple(self.id_stored_entity(*cn).id for cn in asset.iteritems())
        #    #     targets_db.append(entity_ids)
        #    # is this concrete step already known ?
        #    # Look for the sources

        labelnamepairs = list()
        for field, t in zip(assets.target._fields, assets.target):
            if not t._defined:
                # Generate a presumably unique ID
                uniquename = str(uuid.uuid1())
                if isinstance(t, core.File) and t._extension is not None:
                    uniquename += t._extension[0]
                uniquename = os.path.join(absdirname, uniquename)
                # create an entry in the database
                id_storedentity = self._persistent_graph.id_stored_entity(type(t), 
                                                                          uniquename)
                if not id_storedentity.new:
                    # The newly created entity should be... well, NEW.
                    # if reaching here, we have a (serious) problem
                    raise Exception("Bad, bad, bad... the generated unique name %s is not unique." % uniquename)
                t.name = uniquename
                labelnamepairs.append((field, t))
            else:
                # the asset "t" is defined, and sanity checks are (should be?) in self.id_stepconcrete()
                pass
        self._persistent_graph._insert_stepconcrete2storedentities(labelnamepairs, 
                                                                   'target', 
                                                                   stepconcrete_id.id)
        self._persistent_graph.connection.commit()

        sources = list()
        for x in assets.source:
            if x is not None:
                for y in x:
                    sources.append(y)

        self._update_graph(id_stepvariant, stepconcrete_id.id, 
                           tuple(sources),
                           tuple(y for x in assets.target for y in x))
        # store in the cache
        self._cache_dbid[task_hashdb] = stepconcrete_id.id
        return stepconcrete_id

    def _update_graph(self, id_stepvariant, stepconcrete_id, sources, targets):
        # update the graph
        dag = self._dag

        v1 = '%s-%i' % (id_stepvariant, stepconcrete_id)
        if v1 not in dag:
            dag.add_node(v1)
        # add step details (sources and targets)
        steps_before = set()
        for src in sources:
            assert src is not None
            if src not in dag:
                dag.add_node(src)
            vertex = dag[src]
            # edge between the source file and the step
            dag.add_edge(src, v1)
        steps_after = set()
        for target in targets:
            assert target is not None
            if target not in dag:
                dag.add_node(target)
            vertex = dag[target]
            # edge between the step and the target file
            dag.add_edge(v1, target)

    def stepcrawler(self):
        # This is a very rudimentary way to organise tasks. 
        # Bulk synchronous parallel approachs (Google's PREGEL, Apache's Giraf and Hama, Stanford's GPS...)
        # should be considered if ever becoming a large graph.
        dag = self._dag
        # iterate through connected components
        undag = dag.to_undirected()
        for cg in networkx.connected_components(undag):
            # start from a root
            topo = networkx.topological_sort(cg)
            root = topo[0]
            # depth-first search (to get the initial products ASAP)
            for edge in dag.dfs_edges(root):
                stepnode = edge[1]
                useroot_to_do = list()
                # here the `todo` object contains a list of "immediately next"
                # steps to be performed for that one root.
                # Each step can be computed independently

                #FIXME: shouldn't the logic below be taken out of the crawler,
                #       or in a callback function ?
                for step in root.iter_steps():
                    if step.status == _TASK_DONE:
                        continue
                    elif step.status == _TASK_INPROGRESS:
                        # FIXME: have a mechanism to recover from failure, crash, etc... ?
                        continue
                    elif step.status == _TASK_TODO:
                        useroot_to_do.append(step)
                    elif step.status == _TASK_FAILED:
                        raise Exception("Step previously failed: %s" % step)
                    else:
                        # paranoid check
                        raise ValueError("Unknown step %s" % step)
                yield useroot_to_do
                # step in that list should be performed before the next iteration
        pass


    def provenancewalk_storedentity(self, stored_entity,
                                    func_stored_entity,
                                    func_stored_sequence,
                                    func_step_concrete,
                                    func_storedentity_stepconcrete,
                                    func_stepconcrete_storedentity):
        """ Walk up the path.
        :param stored_entity_id: the stored entity to start from
        :param func_stored_entity: a callback called with each stored entity
        :param func_step_concrete: a callback called with each step concrete
        :param func_storedentity_stepconcrete: a callback called with each link between a stored entity and a step concrete
        :param func_stepconcrete_storedentity: a callback called with each link between a step concrete and a stored entity
        """
        assert hasattr(stored_entity, 'id')
        storedentity_stack = collections.deque()
        storedentity_visited = set()
        step_stack = collections.deque()
        step_visited = set()
        if hasattr(stored_entity, 'iter_storedentities'):
            # sequence:
            for elt in stored_entity.iter_storedentities():
                storedentity_stack.append(elt)
                storedentity_visited.add(elt.id)
                func_stored_entity(elt)
            func_stored_sequence(stored_entity)
        else:
            # not a sequence
            storedentity_stack.append(stored_entity)
            storedentity_visited.add(stored_entity.id)
            func_stored_entity(stored_entity)

        while len(storedentity_stack) > 0:
            stored_entity = storedentity_stack.popleft()
            if hasattr(stored_entity, 'iter_storedentities'):
                # this is a sequence
                for elt in stored_entity.iter_storedentities():
                    storedentity_stack.append(elt)
                    storedentity_visited.add(elt.id)
                    func_stored_entity(elt)
                func_stored_sequence(stored_entity)
                stored_entity = storedentity_stack.popleft()                
            step_concrete = self._persistent_graph.get_parenttask_of_storedentity(stored_entity)
            if step_concrete is not None:
                if step_concrete.id not in step_visited:
                    step_stack.append(step_concrete)
                    func_step_concrete(step_concrete)
                    step_visited.add(step_concrete.id)
            while len(step_stack) > 0:
                step_concrete = step_stack.popleft()
                func_stepconcrete_storedentity(step_concrete, stored_entity)
                storedentities = self._persistent_graph.get_srcassets(step_concrete.id)
                for entity in storedentities:
                    if entity.id not in storedentity_visited:
                        storedentity_stack.append(entity)
                        func_stored_entity(entity)
                        storedentity_visited.add(entity.id)
                        #func_stored_entity(entity)
                    func_storedentity_stepconcrete(entity, step_concrete)


    def _destination_walk(self, storedentity_stack, storedentity_visited,
                          step_stack, step_visited,
                          func_stored_entity,
                          func_step_concrete,
                          func_storedentity_stepconcrete,
                          func_stepconcrete_storedentity):
        while len(storedentity_stack) > 0:
            stored_entity = storedentity_stack.popleft()
            steps_concrete = self._persistent_graph.get_targetstepconcrete(stored_entity)
            for entity in steps_concrete:
                if entity.id not in step_visited:
                    step_stack.append(entity)
                    step_visited.add(entity.id)
            func_stored_entity(stored_entity)
            while len(step_stack) > 0:
                step_concrete = step_stack.popleft()
                func_step_concrete(step_concrete)
                func_stepconcrete_storedentity(step_concrete, stored_entity)
                storedentities = self._persistent_graph.get_targetassets(step_concrete.id)
                for entity in storedentities:
                    if entity.id not in storedentity_visited:
                        storedentity_stack.append(entity)
                        storedentity_visited.add(entity.id)
                    func_storedentity_stepconcrete(entity, step_concrete)
        return

    def destinationwalk_stepconcrete(self, step_concrete_id,
                                     func_stored_entity,
                                     func_stored_sequence,
                                     func_step_concrete,
                                     func_storedentity_stepconcrete,
                                     func_stepconcrete_storedentity):
        """ Walk down the path."""
        storedentity_stack = collections.deque()
        storedentity_visited = set()
        step_stack = collections.deque()
        step_visited = set()

        #FIXME: clean the code by taking the nested 'while' loop inside-out
        step_concrete = self._persistent_graph._get_stepconcrete(DbID(step_concrete_id, 
                                                                      False))
        func_step_concrete(step_concrete)
        step_visited.add(step_concrete_id)
        stored_entities = self._persistent_graph.get_targetassets(step_concrete_id)
        for entity in stored_entities:
            if entity.id not in storedentity_visited:
                storedentity_stack.append(entity)
                storedentity_visited.add(entity.id)
            else:
                raise Exception('Step %s has several identical targets.' % str(step_concrete_id))
            #func_stepconcrete_storedentity(step_concrete, entity)
            func_storedentity_stepconcrete(entity, step_concrete)
        return self._destination_walk(storedentity_stack, storedentity_visited,
                                      step_stack, step_visited,
                                      func_stored_entity,
                                      func_step_concrete,
                                      func_storedentity_stepconcrete,
                                      func_stepconcrete_storedentity)

    def destinationwalk_storedentity(self, 
                                     stored_entity,
                                     func_stored_entity,
                                     func_stored_sequence,
                                     func_step_concrete,
                                     func_storedentity_stepconcrete,
                                     func_stepconcrete_storedentity):
        """ Walk down the path."""
        storedentity_stack = collections.deque()
        storedentity_visited = set()
        step_stack = collections.deque()
        step_visited = set()
        
        storedentity_stack.append(stored_entity)
        storedentity_visited.add(stored_entity.id)
        return self._destination_walk(storedentity_stack, storedentity_visited,
                                      step_stack, step_visited,
                                      func_stored_entity,
                                      func_step_concrete,
                                      func_storedentity_stepconcrete,
                                      func_stepconcrete_storedentity)

    def _graph_storedentity(self, db_id,
                            func, opposite,
                            display = GRAPHDISPLAY.STEPLEVEL):
        """Make a provenance graph."""
        se_nodeid = display.value['kwargs']['se_nodeid']
        se_nodelabel = display.value['kwargs']['se_nodelabel']
        step_nodeid = display.value['kwargs']['step_nodeid']
        step_nodelabel = display.value['kwargs']['step_nodelabel']
        def gst(stored_entity, nodelabel):
            return nodelabel % dict((x, getattr(stored_entity, x)) for x in ('id', 'clsname', 'label'))
        def gsc(task, nodelabel):
            return nodelabel % dict((x, getattr(task, x)) for x in ('id', 'clsname'))
        dag = networkx.DiGraph()
        def func_stored_entity(stored_entity):
            dag.add_node(gst(stored_entity, se_nodeid),
                         label=gst(stored_entity, se_nodelabel),
                         shape='record')
        def func_stored_sequence(stored_sequence):
            #dag.add_node(gst(stored_sequence, se_nodelabel),
            #             label=gst(stored_sequence, se_nodelabel),
            #             shape='record')
            for se in stored_sequence.iter_storedentities():
                dag.add_edge(gst(se, se_nodeid),
                             gst(stored_sequence, se_nodeid))                
        def func_step_concrete(step_concrete):
            if step_concrete.status == _TASK_STATUS_LIST[_TASK_DONE]:
                style = 'filled'
            else:
                style = ''
            dag.add_node(gsc(step_concrete, step_nodeid),
                         label=gsc(step_concrete, step_nodelabel),
                         shape='box', 
                         style = style,
                         xlabel=str(step_concrete.status))
        def func_storedentity_stepconcrete(stored_entity, step_concrete):
            if opposite:
                dag.add_edge(gsc(step_concrete, step_nodeid),
                             gst(stored_entity, se_nodeid))                             
            else:
                dag.add_edge(gst(stored_entity, se_nodeid), 
                             gsc(step_concrete, step_nodeid))
        def func_stepconcrete_storedentity(step_concrete, stored_entity):
            if opposite:
                dag.add_edge(gst(stored_entity, se_nodeid),
                             gsc(step_concrete, step_nodeid))                             
            else:
                dag.add_edge(gsc(step_concrete, step_nodeid), 
                             gst(stored_entity, se_nodeid))

        func(db_id,
             func_stored_entity,
             func_stored_sequence,
             func_step_concrete,
             func_storedentity_stepconcrete,
             func_stepconcrete_storedentity)
        return dag

    def provenancegraph_storedentity(self, stored_entity, 
                                     display = GRAPHDISPLAY.STEPLEVEL):
        return self._graph_storedentity(stored_entity,
                                        self.provenancewalk_storedentity,
                                        False, # opposite
                                        display = display)
    def destinationgraph_stepconcrete(self, step_concrete,
                                      display = GRAPHDISPLAY.STEPLEVEL):
        return self._graph_storedentity(step_concrete,
                                        self.destinationwalk_stepconcrete,
                                        True, # opposite
                                        display = display)
    def destinationgraph_storedentity(self, stored_entity,
                                      display = GRAPHDISPLAY.STEPLEVEL):
        return self._graph_storedentity(stored_entity,
                                        self.destinationwalk_storedentity,
                                        True, # opposite
                                        display = display)




    def cleantargets_stepconcrete(self, step_concrete_id):
        """ Clean the targets downstream of a task (step_concrete),
        which means erasing the target files and (re)setting the tasks' status
        to 'TO DO'.
        
        :param step_concrete_id: A task
        :type step_concrete_if: a :class:`DbID` (or anything with an attribute :attr:`id`).
        """
        def func_stored_entity(stored_entity):
            Cls = getattr(self._persistent_graph._model, 
                          stored_entity.clsname)
            instance = Cls(stored_entity.entityname)
            if hasattr(instance, 'iterlistfiles'):
                nameiter = (os.path.join(os.path.dirname(instance.name), basename) \
                            for elt_cls, basename in instance.iterlistfiles())
            else:
                nameiter = iter(instance)
            for pathname in nameiter:
                if os.path.isfile(pathname):
                    os.remove(pathname)
                elif os.path.isdir(pathname):
                    shutil.rmtree(pathname)
                else:
                    warnings.warn("The pathname '%s' associated with the store entity '%s' cannot be removed." % (pathname, str(stored_entity)))
        def func_step_concrete(step_concrete):
            self._persistent_graph.step_concrete_state(step_concrete,
                                                       _TASK_STATUS_LIST[_TASK_TODO])
        def func_storedentity_stepconcrete(stored_entity, step_concrete):
            pass
        def func_stepconcrete_storedentity(step_concrete, stored_entity):
            pass
        def func_stored_sequence(step_concrete, stored_entity):
            pass

        self.destinationwalk_stepconcrete(
            step_concrete_id.id,
            func_stored_entity,
            func_stored_sequence,
            func_step_concrete,
            func_storedentity_stepconcrete,
            func_stepconcrete_storedentity)
    

import sqlite3
import os

sql_fn = os.path.join(os.path.dirname(__file__), 'cache.sql')

class StepConcrete_DbEntry(object):
    __slots__ = ['id', 'status', 'step_variant_id', 'steptype_id', 'executable', 'clsname', 'version', 'parameters']
    def __init__(self, id, status, step_variant_id, 
                 steptype_id, executable, clsname, version, 
                 parameters=()):
        self.id = id
        self.status = status
        self.step_variant_id = step_variant_id
        self.steptype_id = steptype_id
        self.executable = executable
        self.clsname = clsname
        self.version = version
        self.parameters = parameters

TaskStatusCount = namedtuple('TaskStatus', 'id label count')

class StoredEntity(object):
    __slots__ = ('id', 'label', 'clsname', 'entityname')
    def __init__(self, id, label, clsname, entityname):
        self.id = id
        self.label = label
        self.clsname = clsname
        self.entityname = entityname
    def resurrect(self, model):
        if self.clsname == 'NoneType':
            return None
        elif '.' in self.clsname:
            module_name, cls_name = self.clsname.rsplit('.', 1)
            module = import_module(module_name)
        else:
            module = model
            cls_name = self.clsname
        Cls = getattr(module, cls_name)
        return Cls(self.entityname)
    def __str__(self):
        return '\n'.join((super(StoredEntity, self).__str__(),
                          'id: %i' % self.id))

class StoredSequence(object):
    __slots__ = ('id', 'label', 'clsname', 'storedentities')
    def __init__(self, id, label, clsname, storedentities):
        self.id = id
        self.label = label
        self.clsname = clsname
        self.storedentities = storedentities
    def resurrect(self, model):
        if '.' in self.clsname:
            module_name, cls_name = self.clsname.rsplit('.', 1)
            module = import_module(module_name)
        else:
            module = model
            cls_name = self.clsname
        Cls = getattr(module, cls_name)
        assert issubclass(Cls, core.FileSequence), 'Expected a FileSequence but got %s' % str(Cls)
        return Cls(x.resurrect(model) for x in self.storedentities)
    def iter_storedentities(self):
        return iter(self.storedentities)
    def __iter__(self):
        return self.iter_storedentities()
    def __len__(self):
        return len(self.storedentities)
    def __str__(self):
        return '\n'.join((super(StoredSequence, self).__str__(),
                          'id: %i' % self.id,
                          'length: %i' % len(self)))


class PersistentTaskGraph(object):
    """
    List of tasks stored on disk.
    
    """

    StoredEntityNoLabel = namedtuple('StoredEntityNoLabel', 'id clsname entityname')

    def __init__(self, db_fn, model, wd='.', force_create=False, isolation_level=None):
        """
        :param db_fn: file name for the database.
        :param wd: working directory (where derived data files are stored)
        :param force_create: recreate database if a file with the same name is already
                             present.
        :param isolation_level: passed to :func:`sqlite3.connection`.
        """

        # check whether the file containing the SQLite database exists.
        # 
        if os.path.exists(db_fn):
            if os.path.isfile(db_fn):
                if force_create:
                    create = True
                else:
                    create = False
            else:
                raise ValueError('db_fn should be a file name, not a directory name.')
        else:
            create = True
        self._db_fn = db_fn
        self._model = model
        self._model_steplist = unifex._make_stepdict(model)
        self._wd = wd
        self.created = create
        connection = sqlite3.connect(db_fn, isolation_level=isolation_level)
        self.connection = connection
        if create:
            self._create(db_fn, wd)
        self._fill_tasklist()
        self._statuslist = None # lazy evaluation

    def _create(self, db_fn, wd):
        with open(sql_fn, 'r') as fh:
            sql = fh.readlines()
        sql = ''.join(sql)
        connection = self.connection
        with connection:
            for block in sql.split(';'):
                connection.execute(block)
        # misc
        #    version
        sql = """
        INSERT INTO misc (tag, description, val) VALUES (?, ?, ?)
        """
        with connection:
            res = connection.executemany(sql, 
                                         (('version','Version for the framework',
                                           __version__),
                                          ('workingdirectory', 'Directory in which files are saved',
                                           wd)))

        # populate possible status for tasks
        sql = """
        INSERT INTO step_status (label) VALUES (?)
        """
        with connection:
            res = connection.executemany(sql, ((x,) for x in _TASK_STATUS_LIST.keys()))

    def _fill_tasklist(self):
        cursor = self.connection.cursor()
        sql = """
        SELECT id, label
        FROM step_status
        """
        res = cursor.execute(sql)
        for x in res:
            _TASK_STATUS_LIST[x[1]] = x[0]


    def getversion(self):
        sql = """
        SELECT val 
        FROM misc
        WHERE tag=='version'
        """
        cursor = self.connection.cursor()
        res = cursor.execute(sql)
        return res.fetchone()[0]
    version = property(getversion, None, None,
                       "Version for the database and package (mixing versions comes at one's own risks)")

    def iter_steps(self):
        """ Iterate through the concrete steps """
        sql = """
        SELECT step_concrete.id AS id,
               step_status.label AS status,
               sv.id,
               sv.step_type_id,
               sv.executable,
               sv.cls,
               sv.version
        FROM step_concrete, step_status, step_variant AS sv
        WHERE step_concrete.step_status_id=step_status.id
        AND step_concrete.step_variant_id=sv.id
        ORDER BY step_concrete.id -- ID is autoincremented, so this is like a chronological order
        """
        # If the SQL above changes, the definition of
        # StepConcrete_DbEntry might have to be updated
        cursor = self.connection.cursor()
        steps = cursor.execute(sql)
        # loop through the steps
        for row in steps:
            # for each step, get the sources
            sc = StepConcrete_DbEntry(*row)
            sources = tuple(self.get_srcassets(sc.id))
            targets = tuple(self.get_targetassets(sc.id))
            parameters = tuple(self.get_parameters(sc.id))
            yield Step(sc, sources, targets, parameters, self._model)

    def _get_assets(self, concrete_step, kind):
        if hasattr(concrete_step, 'id'):
            concrete_step_id = concrete_step.id
        else:
            concrete_step_id = int(concrete_step)
        cursor = self.connection.cursor()
        # First yield stored entities
        sql_template = """
        SELECT stored_entity.id as id,
               step_concrete2%(kind)sfile.label as label,
               stored_entity.classname as classname,
               stored_entity.entityname as entityname
        FROM step_concrete2%(kind)sfile, stored_entity
        WHERE step_concrete_id=?
        AND stored_entity_id=stored_entity.id
        """
        sql = sql_template % {'kind': kind}
        res = cursor.execute(sql, (concrete_step_id,))
        for src in res:
            res = StoredEntity(*src)
            yield res
        # Now yield sequences (of stored entities)
        sql_template_ss = """
        SELECT stored_sequence.id,
               sc.label as label,
               stored_sequence.classname as classname
        FROM step_concrete2%(kind)sfile AS sc
        JOIN stored_sequence
            ON stored_sequence.id=sc.stored_sequence_id
        WHERE step_concrete_id=?
        """
        sql = sql_template_ss % {'kind': kind}
        sql_se = """
        SELECT stored_entity.id as id,
               stored_entity.classname as classname,
               stored_entity.entityname as entityname,
               pos
        FROM stored_entity2sequence AS se2s
        JOIN stored_entity
            ON se2s.stored_entity_id=stored_entity.id
        WHERE se2s.stored_sequence_id=?
        ORDER BY id, classname, pos
        """
        res = cursor.execute(sql, (concrete_step_id,))
        cursor2 = self.connection.cursor()
        for src in res:
            res2 = cursor2.execute(sql_se, (src[0],))
            ses = list()
            for pos, x in enumerate(res2.fetchall()):
                if x[-1] != pos:
                    raise Exception('The sequence is missing an item in postion %i (and got position %i instead).' % (pos, x[-1]))            
                ses.append(StoredEntity(x[0], None, x[1], x[2]))
            res = StoredSequence(src[0], src[1], src[2], ses)
            yield res


    def task_time_points(self, task_id):
        cursor = self.connection.cursor()
        sql = """
        SELECT time_creation,
               time_t0,
               time_t1
        FROM step_concrete
        WHERE step_concrete.id=?
        """
        cursor.execute(sql, (task_id.id, ))
        res = cursor.fetchone()
        return res

    def set_time_t0(self, task_id, t0):
        cursor = self.connection.cursor()

        sql = """
        UPDATE step_concrete
        SET time_t0=?
        WHERE step_concrete.id=?
        """
        cursor.execute(sql, (t0, task_id.id, ))
        self.connection.commit()

    def set_time_t1(self, task_id, t1):
        cursor = self.connection.cursor()

        sql = """
        UPDATE step_concrete
        SET time_t1=?
        WHERE step_concrete.id=?
        """
        cursor.execute(sql, (t1, task_id.id, ))
        self.connection.commit()
        
    def _get_stepconcrete(self, step_concrete_id):
        cursor = self.connection.cursor()
        sql = """
        SELECT step_concrete.id,
               step_concrete.step_status_id,
               step_concrete.step_variant_id,
               step_type_id,
               executable,
               cls,
               version,
               step_parameters.json
        FROM step_concrete
        INNER JOIN step_status
        ON step_concrete.step_status_id=step_status.id
        INNER JOIN step_variant AS sv
        ON step_concrete.step_variant_id=sv.id
        INNER JOIN step_concrete2parameters AS sc2p
        ON sc2p.step_concrete_id=step_concrete.id
        INNER JOIN step_parameters
        ON step_parameters_id=step_parameters.id
        WHERE step_concrete.id=?
        """
        cursor.execute(sql, (step_concrete_id.id, ))
        row = cursor.fetchone()
        if row is None:
            # Trouble. Trying to provide an informative error message
            sql = """
            SELECT *
            FROM step_concrete
            WHERE id=?
            """
            cursor.execute(sql)
            if len(cursor.fetchall()) == 0:
                raise ValueError("The task ID %i cannot be found." % step_concrete_id.id)
            else:
                raise ValueError("Error with the database. Likely because of interrupted process while entering tasks.")

        res = StepConcrete_DbEntry(*row)
        return res



    _SQL_TASK_FROMASSET_TEMPLATE = """
        SELECT step_concrete.id,
               step_concrete.step_status_id,
               step_concrete.step_variant_id,
               step_type_id,
               executable,
               cls,
               version,
               step_parameters.json
        FROM step_concrete2%(src_or_target)sfile AS sc2f
        INNER JOIN step_concrete
        ON sc2f.step_concrete_id=step_concrete.id
        INNER JOIN step_variant
        ON step_variant_id=step_variant.id
        INNER JOIN step_concrete2parameters AS sc2p
        ON sc2p.step_concrete_id=step_concrete.id
        INNER JOIN step_parameters
        ON step_parameters_id=step_parameters.id
        WHERE sc2f.stored_%(entity_or_sequence)s_id=?
    """
    _SQL_TASK_TOTARGETSEQUENCE = _SQL_TASK_FROMASSET_TEMPLATE  % {'src_or_target': 'target',
                                                                  'entity_or_sequence': 'sequence'}
    _SQL_TASK_FROMSRCSEQUENCE  = _SQL_TASK_FROMASSET_TEMPLATE % {'src_or_target': 'src',
                                                                 'entity_or_sequence': 'sequence'}
    _SQL_TASK_TOTARGETSENTITY  = _SQL_TASK_FROMASSET_TEMPLATE % {'src_or_target': 'target',
                                                                 'entity_or_sequence': 'entity'}
    _SQL_TASK_FROMSRCSENTITY   = _SQL_TASK_FROMASSET_TEMPLATE % {'src_or_target': 'src',
                                                                 'entity_or_sequence': 'entity'}

    def _get_stepconcrete_from_storedentity(self, stored_entity_id, sql):
        """ 
        Retrieve the step_concrete object(s) linked with a stored_entity_id.
        Whenever :param:`kind` is equal to "target" there will be at most
        one stored_entity (if there is more one, the database has a consistentcy
        issue).
        :param stored_entity_id: the ID for a stored_entity
        :param kind: either 'src' or 'target'
        """
        cursor = self.connection.cursor()
        res = cursor.execute(sql, (stored_entity_id,))
        for src in res:
            res = StepConcrete_DbEntry(*src)
            yield res

    def get_parenttask_of_storedentity(self, stored_entity):
        """ Return the task producing a stored entity. There should obviously only be one such task,
        and an Exception is raised if not the case.
        :param stored_entity: the stored entity in the database
        :type stored_entity: :class:`StoredEntity`
        :rtype: a `StepConcrete_DbEntry` :class:`namedtuple`, or None
        """
        assert hasattr(stored_entity, 'id')
        step_concrete = None
        if hasattr(stored_entity, 'iter_storedentities'):
            raise ValueError('Only atomic stored entities are accepted.')
        else:
            sql = self._SQL_TASK_TOTARGETSENTITY
        for i, step_concrete in enumerate(self._get_stepconcrete_from_storedentity(stored_entity.id, sql)):
            if i > 0:
                raise Exception("""
 Consistency issue with the database. More than one step is claiming to be the source of the stored_entity_id "%s" """ % str(stored_entity.id))
        return step_concrete

    def get_targetstepconcrete(self, stored_entity):
        """ Return the tasks using a given stored entity.
        :param stored_entity: the stored entity in the database.
        :type stored_entity: can be :class:`StoredEntity` or :class:`StoredSequence`
        :rtype: a `SepConcrete_DbEntry` :class:`namedtuple`, or None
        """
        if hasattr(stored_entity, 'iter_storedentities'):
            raise NotImplementedError('Handling sequence as source assets is not yet implemented.')
        else:
            sql = self._SQL_TASK_FROMSRCSENTITY

        return tuple(self._get_stepconcrete_from_storedentity(stored_entity.id, 
                                                              sql))
        
    def get_srcassets(self, concrete_step_id):
        """ Return the source files for a given concrete step ID.
        :param concrete_step_id: ID for the concrete step in the database.
        :rtype: generator
        """
        return self._get_assets(concrete_step_id, 'src')

    def get_targetassets(self, concrete_step_id):
        """ Return the target files for a given concrete step ID.
        :param concrete_step_id: ID for the concrete step in the database.
        :rtype: generator
        """
        return self._get_assets(concrete_step_id, 'target')

    def get_parameters(self, concrete_step_id):
        cursor = self.connection.cursor()
        sql = """
        SELECT step_parameters.id as id,
               step_parameters.json as json
        FROM step_concrete2parameters, step_parameters
        WHERE step_concrete_id=?
        AND step_parameters_id=step_parameters.id
        """
        res = cursor.execute(sql, (concrete_step_id,))
        return res.fetchall()

    def get_statuslist(self):
        if self._statuslist is None:
            cursor = self.connection.cursor()
            sql = """
            SELECT * from step_status
            """
            res = cursor.execute(sql)
            self._statuslist = tuple(res)
        return self._statuslist

    statuslist = property(get_statuslist, None, None,
                          """ Status list """)

    def id_stepparameters(self, parameters):
        """
        Conditionally add parameters (add only if not already present)
        :param parameters: sequence of parameters
        :rtype: ID for the pattern as a :class:`DbID`.
        """
        cursor = self.connection.cursor()
        param_json = json.dumps(parameters)
        sql = """
        SELECT id 
        FROM step_parameters
        WHERE json=?
        """
        res = cursor.execute(sql, (param_json,))
        res = cursor.fetchone()
        if res is None:
            sql = """
            INSERT INTO step_parameters (
            json
            ) VALUES (
            ?);
            """
            cursor.execute(sql, (param_json,))
            db_id = cursor.lastrowid
            self.connection.commit()
            res = DbID(db_id, True)
        else:
            res = DbID(res[0], False)
        return res

        
    def id_stored_entity(self, cls, name):
        """
        Conditionally add a stored entity (add only if not already present)
        :param cls: Python class for the stored entity
        :param name: Parameter "name" for the class "cls".
        :rtype: ID for the pattern as a :class:`DbID`.
        """
        cursor = self.connection.cursor()
        sql = """
        SELECT id 
        FROM stored_entity
        WHERE classname=?
        AND entityname=?;
        """
        res = cursor.execute(sql, (cls.__name__, name))
        res = cursor.fetchone()
        if res is None:
            sql = """
            INSERT INTO stored_entity (
            classname,
            entityname
            ) VALUES (
            ?,
            ?);
            """
            try:
                cursor.execute(sql, (cls.__name__, name))
            except sqlite3.IntegrityError as ie:
                # print an informative message before propagating the exception
                sql = """
                SELECT classname
                FROM stored_entity
                WHERE entityname=?
                """
                res = cursor.execute(sql, (name,))
                res = cursor.fetchone()
                msg = 'ERROR: The descriptor "%s" for the type "%s" is already in use with type "%s"' % (name, cls.__name__, res[0])
                print(msg)
                raise(ie)
            db_id = cursor.lastrowid
            self.connection.commit()
            res = DbID(db_id, True)
        else:
            res = DbID(res[0], False)
        return res

    def id_stored_sequence(self, cls, clsname_sequence):
        """
        Conditionally add a sequence of stored entities (add only if not already present)
        :param clsname_sequence: Sequence of pairs (Python class for the stored entity, parameter "name" for the class "cls")
        :rtype: ID for the sequence as a :class:`DbID`.
        """
        cursor = self.connection.cursor()

        sql = """
        SELECT stored_sequence_id
        FROM stored_entity2sequence
        WHERE stored_entity_id=?
        AND pos=?;
        """
        stored_entity_ids = list()
        candidates = None
        # iterate through the entities in the sequence
        for pos, (cls_elt, name_elt) in enumerate(clsname_sequence):
            # ensure that the entity is tracked
            se_id = self.id_stored_entity(cls_elt, name_elt)
            stored_entity_ids.append(se_id)
            res = cursor.execute(sql, (se_id.id, pos))
            if candidates is None:
                candidates = set(x[0] for x in cursor.fetchall())
            else:
                candidates = candidates.intersection(set(x[0] for x in cursor.fetchall()))
        if candidates is None or len(candidates) == 0:
            sql = """
            INSERT INTO stored_sequence (
            classname
            ) VALUES (
            ?);
            """
            cursor.execute(sql, (cls.__name__, ))
            db_id = cursor.lastrowid
            # 
            sql = """
            INSERT INTO stored_entity2sequence (
            stored_sequence_id,
            stored_entity_id,
            pos
            ) VALUES (
            ?, ?, ?);
            """
            for pos, (se_id) in enumerate(stored_entity_ids):
                res = cursor.execute(sql, (db_id, se_id.id, pos))
            
            self.connection.commit()
            res = DbID(db_id, True)
        elif len(candidates) == 1:
            res = DbID(next(iter(candidates)), False)
        else:
            raise Exception("Consistency issue with the database.")
        return res

    def id_step_activity(self, activity):
        """ 
        Conditionally add an activity (add only if not already present)
        :param activity: one actibity name
        :rtype: ID for the activity as an integer
        """
        cursor = self.connection.cursor()
        sql = """
        SELECT id
        FROM step_activity
        WHERE label=?;
        """
        res = cursor.execute(sql, (activity.value,))
        res = cursor.fetchone()
        if res is None:
            sql = """
            INSERT INTO step_activity
            (label) VALUES (?);
            """
            cursor.execute(sql, (activity.value,))
            res = cursor.lastrowid
            self.connection.commit()
            res = DbID(res, True)
        else:
            res = DbID(res[0], False)
        return res

    def id_step_type(self, activities):
        """
        Conditionally add a step type (add only if not already present).
        :param activities: sequence of activity names
        :rtype: ID for the step type as an integer
        """

        assert len(activities) > 0, "The number of activities must be > 0"

        cursor = self.connection.cursor()

        # first get the activity IDs
        activity_ids = tuple(self.id_step_activity(x) for x in activities)
        steptype_id = None
        if not any(x.new for x in activity_ids):
            # query if any step_type already has all the associated activities
            # The if statement saves that trouble if any activity had to be created
            # (since then the step_type cannot possibly be already in the database)
            sql_activity = """SELECT step_type_id FROM step_type2activity WHERE step_activity_id = %i"""

            sql = """
            SELECT step_type_id
            FROM
              (SELECT step_type_id
               FROM step_type2activity
               WHERE step_activity_id IN
            """ + \
                '    (' + '\nINTERSECT\n'.join((sql_activity % dbid.id) for dbid in activity_ids) + ')' + \
                """
                EXCEPT
                SELECT step_type_id
                FROM step_type2activity
                WHERE step_activity_id NOT IN (%s)
                )
                """ % ','.join(str(dbid.id) for dbid in activity_ids)
            res = cursor.execute(sql)

            steptype_id = res.fetchall()
            
        if steptype_id is None or len(steptype_id) == 0:
            # if no activity ID, we should create one
            sql = """
            INSERT INTO step_type VALUES(null);
            """
            res = cursor.execute(sql)
            steptype_id = cursor.lastrowid

            sql = """
            INSERT INTO step_type2activity
            (step_type_id, step_activity_id)
            VALUES (?, ?);
            """
        
            for activity_id in activity_ids: 
                cursor.execute(sql, (steptype_id, activity_id.id))
            self.connection.commit()
            return DbID(steptype_id, True)
        elif len(steptype_id) > 1:
            #FIXME: if more than one steptype_id found, we have a problem
            raise Exception("Houston, we have problem. We have several step types (%s) with all the activities." % repr(activities))

        steptype_id = steptype_id[0][0]
        return DbID(steptype_id, False)

    def _iter_step_type(self):
        sql = """
        SELECT step_type.id, activity
        FROM step_type
        INNER JOIN step_type2activity
        ON step_type2activity.step_type_id=step_type.id
        INNER JOIN step_activity
        ON step_type2activity.step_activity_id=step_activity.id
        GROUP BY step_type.id
        """
        cursor = self.connection.cursor()
        cursor.execute(sql)
        while True:
            rows = cursor.fetchmany(100)
            if not rows:
                break
            for row in rows:
                yield row
            

    def id_step_variant(self,
                        step,
                        activities, # list of activities the step is covering (step_id inferred from that)
                    ):
        """
        Return a database ID for the step variant (creating a new ID only of the variant is not already tracked)
        
        :param step: a step
        :type step: :class:`core.StepAbstract`
        :param activities: a sequence of activity names

        :rtype: ID for a step variant as an :class:`int`.
        """

        assert isinstance(step, StepAbstract)

        executable = step._execpath
        version = step.version

        step_type_id = self.id_step_type(activities)

        cursor = self.connection.cursor()
        step_variant_id = None
        if not step_type_id.new:
            # not a new step_id, so may be the step_variant is known as well
            if executable is None:
                sql = """
                SELECT id
                FROM step_variant
                WHERE step_type_id=?
                AND executable IS NULL
                AND cls=?
                AND version=?
                """
                t = type(step)
                sql_params = (step_type_id.id,
                              '.'.join((t.__module__, t.__name__)),
                              version)
            else:
                sql = """
                SELECT id
                FROM step_variant
                WHERE step_type_id=?
                AND executable=?
                AND cls=?
                AND version=?
                """
                t = type(step)
                sql_params = (step_type_id.id,
                              executable,
                              '.'.join((t.__module__, t.__name__)),
                              version)
            res = cursor.execute(sql, sql_params)
            step_variant_id = res.fetchone()
            # FIXME: test if more than one ? 
            if step_variant_id is None:
                step_variant_id = None
            else:
                step_variant_id = step_variant_id[0] # ID only
        if step_variant_id is None:
            sql = """
            INSERT INTO step_variant
            (step_type_id, executable, cls, version)
            VALUES
            (?, ?, ?, ?)
            """
            t = type(step)
            res= cursor.execute(sql, (step_type_id.id,
                                      executable,
                                      '.'.join((t.__module__, t.__name__)),
                                      version))
            step_variant_id = cursor.lastrowid
            self.connection.commit()
            return DbID(step_variant_id, True)
        else:
            return DbID(step_variant_id, False)


    @property
    def nconcrete_steps(self):
        cursor = self.connection.cursor()
        sql = """
        SELECT COUNT(id)
        FROM step_concrete
        """
        res = cursor.execute(sql)
        return res.fetchone()[0]

    @property
    def nconcrete_steps_status(self):
        cursor = self.connection.cursor()
        sql = """
        SELECT step_status.id, step_status.label, COUNT(*)
        FROM step_concrete,
             step_status
        WHERE step_concrete.step_status_id=step_status.id
        GROUP BY step_status.id

        """
        res = cursor.execute(sql)
        return tuple(TaskStatusCount(*x) for x in res.fetchall())

    def id_stepconcrete(self, step_variant_id,
                        sources, targets, parameters,
                        tag = 1):
        """
        Conditionally add a task ("concrete" step),
        that is a step variant (executable and parameters)
        to which source and target files, as well as parameters, are added.

        :param step_variant_id: ID for the step variant
        :type step_variant_id: integer
        :param sources: sequence of sources
        :type sources: :class:`AssetSet`
        :param targets: sequence of targets
        :type targets: :class:`AssetSet`
        :param parameters: list of parameters
        :type parameters: a sequence of :class:`str`
        :param tag: a tag, used to performed repetitions of the exact same task
        :type tag: a sequence of :class:`int`

        :rtype: :class:`DbID`

        """

        assert isinstance(step_variant_id, int)

        cursor = self.connection.cursor()

        # Initialize at step_concrete_id to None. If still None later,
        # it will mean that there is no matching task already tracked.
        step_concrete_id = None 
        #FIXME: transaction (in case one of the inserts fails)
        # DB IDs for all sources

        # ID for the parameters
        parameters_id = self.id_stepparameters(parameters)

        if len(sources) == 0:
            # special case: no sources
            # Query whether the task is already known
            sql = """
            SELECT id,
                   step_variant_id,
                   step_status_id
            FROM step_concrete
            WHERE step_variant_id=?
            AND tag=?
            AND id NOT IN (SELECT id FROM step_concrete2srcfile)
            GROUP BY step_concrete.id
            """
            cursor.execute(sql, (step_variant_id, tag))
            res = cursor.fetchall()
            if len(res) == 1:
                step_concrete_id = res[0][0]
            elif len(res) == 0:
                # The task is not known (step_concrete_id is already set to None)
                pass
            elif len(res) > 1:
                # Several tasks are matching. This is not good.
                raise Exception("Serious trouble.")
            else:
                raise Exception("We should never be here.")
        else:
            # Query whether the task is already known. Information about the provenance
            # (including parameters) is enough.
            sources_db = list()
            sql_template = """
            SELECT step_concrete.id,
                   step_variant_id,
                   step_status_id,
                   stored_entity_id,
                   stored_sequence_id
            FROM step_concrete,
                 step_concrete2srcfile,
                 step_concrete2parameters
            WHERE step_variant_id=?
            AND step_parameters_id=?
            AND label=?
            AND step_concrete2parameters.step_concrete_id=step_concrete.id
            AND step_concrete2srcfile.step_concrete_id=step_concrete.id
            AND stored_entity_id%(se)s
            AND stored_sequence_id%(ss)s
            AND step_concrete.tag=?
            """
            sql_se = sql_template % {'se': '=?', 'ss': ' IS NULL'}
            sql_ss = sql_template % {'ss': '=?', 'se': ' IS NULL'}
            candidates = None # candidate known tasks
            for (label, asset, assetattr) in zip(sources._fields, sources, sources._sources):
                # loop over the source assets
                if asset is None and assetattr.allownone:
                    continue
                asset_items = tuple(cn for cn in asset.iteritems())
                if isinstance(asset, core.FileSequence):
                    # if a sequence we also need to ensure that the sequence itself is tracked
                    sequence_id = self.id_stored_sequence(type(asset), asset.iteritems()).id
                    se_id = None
                    etid = (step_variant_id, 
                            parameters_id.id,
                            label,
                            #se_id,
                            sequence_id,
                            tag)
                    sql = sql_ss
                else:
                    # if not a sequence we need to ensure that object (file) is tracked
                    sequence_id = None
                    for se_i, cn in enumerate(asset.iteritems()):
                        if se_i > 0:
                            # we should never reach here, because it was tested above
                            raise Exception("Only FileSequence objects should have several saved entities.")
                        # ensure the that the stored ID for this asset is tracked in the DB
                        se_id = self.id_stored_entity(*cn).id
                    # create a tuple of parameters for the query that selects stored entities.
                    etid = (step_variant_id, 
                            parameters_id.id,
                            label,
                            se_id,
                            #sequence_id,
                            tag)
                    sql = sql_se
                # query with parameters
                cursor.execute(sql, etid)
                # retrieve the task_id using this source asset
                tmp = set(x[0] for x in cursor.fetchall())
                if candidates is None:
                    # first iteration, set is whatever is retrieved
                    candidates = tmp
                else:
                    # otherwise, the candidate task_ids are the intersection of candidates so far and what was returned 
                    candidates = candidates.intersection(tmp)
                if len(candidates) == 0:
                    # no candidate left: the task is not yet tracked.
                    break
            if len(candidates) > 1:
                raise Exception("DB consistency error: several candidate tasks were found.")
            elif len(candidates) == 1:
                step_concrete_id = next(iter(candidates))
            else:
                # step_concrete_id remains None
                if step_concrete_id is not None:
                    raise Exception('')

        if step_concrete_id is not None:
            # the concrete step is already in the database
            res = DbID(step_concrete_id, False)
            # The following is a check of the targets. 
            target_entities = tuple(self.get_targetassets(res))
            if len(target_entities) != len(targets):
                raise ValueError(''.join(('The task was already stored but with a different number of targets ',
                                          '(%i, but now %i).' % (len(target_entities), len(targets)))))
                
            for entity in target_entities:
                targetasset = getattr(targets, entity.label)
                if targetasset._defined:
                    # Check that this is matching the same as what is in the database
                    # Being thorough is relatively important here
                    if isinstance(entity, StoredSequence):
                        raise NotImplementedError('Sequence assets in targets is not yet handled.')
                    # check that constructor's attribute is the same:
                    if targetasset.name != entity.entityname or type(targetasset).__name__ != entity.clsname:
                        raise ValueError(''.join(('The task was already stored but with different values for the target "%s"' % entity.label,
                                                  ' (%s(%s) instead of %s(%s)).' % (entity.clsname, entity.entityname,
                                                                                    type(targetasset).__name__, targetasset.name))))
                else:
                    if isinstance(entity, StoredSequence):
                        #
                        raise NotImplementedError('Sequence assets in targets is not yet handled.')
                    else:
                        # set it to what is in the database
                        targetasset.name = entity.entityname
        else:
            # the concrete step was never seen before
            param_json = json.dumps(parameters_id)
            res = self._add_stepconcrete(step_variant_id,
                                         sources,
                                         targets,
                                         parameters_id,
                                         tag = tag)
        return res
        
    def _add_stepconcrete(self, step_variant_id,
                          sources,
                          targets,
                          parameters_id,
                          tag = 1):
        """
        Add a "concrete" step, that is a step variant (executable and parameters)
        to which source and target files are added.

        :param step_variant_id: ID for the step variant
        :type step_variant_id: integer
        :param sources: sequence of sources
        :type sources: :class:`AssetSet`
        :param targets: sequence of targets
        :type targets: :class:`AssetSet`
        :param parameters_id: ID for the parameters
        :type tag: A tag for the task
        :param tag: :class:`int`
        :rtype: :class:`DbID`

        """

        assert isinstance(step_variant_id, int)

        cursor = self.connection.cursor()
        #FIXME: transaction (in case one of the inserts fails) only committed
        #       before exiting this method, but can this be bypassed by SQLite global settings ?
        #       (I am relying on the current default)
                
        sql = """
        INSERT INTO step_concrete (
        step_variant_id,
        step_status_id,
        time_creation,
        tag
        ) VALUES (
        ?,
        ?,
        ?,
        ?);
        """
        # last update time set to NULL

        #FIXME: try/catch. since step_variant_id is a foreign key,
        #       this will fail if there is no such step_variant_id
        cursor.execute(sql, (step_variant_id, _TASK_STATUS_LIST[_TASK_TODO], time.time(), tag))
        step_concrete_id = DbID(cursor.lastrowid, True)

        labelnamepairs = zip(sources._fields, sources)
        self._insert_stepconcrete2storedentities(labelnamepairs, 'src', step_concrete_id.id)

        # ensure that defined target are tracked
        labelnamepairs = list()
        for field, t in zip(targets._fields, targets):
            if t._defined:
                labelnamepairs.append((field, t))
        self._insert_stepconcrete2storedentities(labelnamepairs, 'target', step_concrete_id.id)
        self.connection.commit()

        self._insert_stepconcrete2parameters(parameters_id.id, step_concrete_id.id)
        self.connection.commit()
        return step_concrete_id
        #return Step(step_concrete_id, sources, targets)

    def _insert_stepconcrete2parameters(self, parameters_id, step_concrete_id):
        """ WARNING: add a commit after the call !!!"""
        cursor = self.connection.cursor()

        sql = """
        INSERT INTO step_concrete2parameters (
        step_concrete_id,
        step_parameters_id
        ) VALUES (
        ?,
        ?);
        """

        cursor.execute(sql,
                       (step_concrete_id, parameters_id))

    def _insert_stepconcrete2storedentities(self, labelnamepairs, what, step_concrete_id):
        """ WARNING: add a commit in caller after this returns !!!"""
        assert what in ('src', 'target')
        cursor = self.connection.cursor()

        sql_template = """
        INSERT INTO step_concrete2%(what)sfile (
        label,
        step_concrete_id,
        stored_%(the)s_id
        ) VALUES (
        ?,
        ?,
        ?);
        """
        # 2 SQL query: one for assets that are sequences and one for assets that are scalars.
        sql_ss = sql_template % {'what': what, 'the': 'sequence'}
        sql_se = sql_template % {'what': what, 'the': 'entity'}

        for (label, asset) in labelnamepairs:
            if isinstance(asset, core.FileSequence):
                # if a sequence we also need to create the sequence itself
                sequence_id = self.id_stored_sequence(type(asset), asset.iteritems()).id
                se_id = None
                etid = (label, 
                        step_concrete_id,
                        # stored_entity
                        sequence_id)
                sql = sql_ss
            else:
                sequence_id = None
                if asset is None:
                    # if the asset is not defined here, 
                    # it can / should only be the case
                    # because it is allowed to be None
                    continue
                    #se_id = self.id_stored_entity(type(asset), None).id
                else:
                    for se_i, cn in enumerate(asset.iteritems()):
                        if se_i > 0:
                            raise Exception("Only FileSequence object should have several saved entities.")
                        # ensure the that the stored IDs are tracked in the DB
                        se_id = self.id_stored_entity(*cn).id
                etid = (label, 
                        step_concrete_id, 
                        se_id
                        # stored_sequence
                )
                sql = sql_se
            res = cursor.execute(sql, etid)
            
        #labelstepandnames = tuple( for cn in asset.iteritems())
        #cursor.executemany(sql,
        #labelstepandnames)

    def step_concrete_info(self, step_concrete_id):
        sql = """
        SELECT
          step_variant_id,
          step_status.label,
          time_creation
        FROM  step_concrete, step_status
        WHERE step_concrete.id=?
        AND step_status.id = step_status_id
        """
        cursor = self.connection.cursor()
        cursor.execute(sql, (step_concrete_id.id, ))
        res = cursor.fetchone()
        return res

    def step_concrete_state(self, step_concrete_id,
                            state_id):
        """ Set the state of a task:

        - step_concrete_id: task ID (DbID)
        - state_id:         state ID
        """
        sql = """
        UPDATE step_concrete
        SET step_status_id=?
        WHERE step_concrete.id=?
        """
        cursor = self.connection.cursor()
        res = cursor.execute(sql, (state_id, step_concrete_id.id))
        self.connection.commit()
        #FIXME: return anything ?

    def step_concrete_status(self, step_concrete_id):
        sql = """
        SELECT step_status.id, step_status.label
        FROM step_concrete
        INNER JOIN step_status
        ON step_status_id=step_status.id
        WHERE step_concrete.id=?
        """
        cursor = self.connection.cursor()
        cursor.execute(sql, (step_concrete_id.id,))
        return cursor.fetchall()

    def _get_assetsofactivity(self, activity, what):
        """ 
        Find the assets associated with steps performing a specific activity.
        :param activity: an activity
        :type activity: :class:`Enum`
        """
        cursor = self.connection.cursor()

        # -- find step type IDs performing a specific activity
        sql = """
        SELECT DISTINCT step_type_id
        FROM step_activity, step_type2activity
        WHERE label=?
        AND step_activity.id=step_type2activity.step_activity_id
        """
        res = cursor.execute(sql, (activity.value,))
        #FIXME: check that activity with that label found ?
        step_type_ids = res.fetchall()

        # -- find concrete step IDs with that activity
        sql = """
        SELECT step_concrete.id
        FROM step_variant, 
             step_concrete,
             step_concrete2%(what)sfile
        WHERE step_type_id=?
        AND step_variant.id=step_variant_id
        AND step_concrete.id=step_concrete2%(what)sfile.step_concrete_id
        """ % {'what': what}
        step_concrete_ids = list()
        for step_type_id in step_type_ids:
            res = cursor.execute(sql, step_type_id)
            step_concrete_ids.extend(res.fetchall())

        # -- find the stored entities resulting from the concrete steps above
        results = list()
        for task_dbentry in step_concrete_ids:
            res = self.get_targetassets(task_dbentry[0])
            results.extend(res)

        # -- 
        
        return tuple(results)
        
    def get_sourcesofactivity(self, activity):
        """ 
        Retrieve the sources of steps performing a specific activity.
        :param activity: an activity
        :type activity: :class:`Enum`
        """
        return self._get_assetsofactivity(activity, 'src')

    def find_targetsofactivity(self, activity):
        warnings.warn('find_targetsofactivity() is deprecated, use get_targetsofactivity().')
        return self._get_assetsofactivity(activity, 'target')

    def get_targetsofactivity(self, activity):
        """ 
        Retrieve the targets of steps performing a specific activity.
        :param activity: an activity
        :type activity: :class:`Enum`
        """
        cursor = self.connection.cursor()

        # -- find step type IDs performing a specific activity
        sql = """
        SELECT DISTINCT step_type_id
        FROM step_activity, step_type2activity
        WHERE label=?
        AND step_activity.id=step_type2activity.step_activity_id
        """
        res = cursor.execute(sql, (activity.value,))
        step_type_ids = res.fetchall()
        if len(step_type_ids) == 0:
            sql = """
            SELECT DISTINCT id
            FROM step_activity
            WHERE label=?
            """
            res = cursor.execute(sql, (activity.value,))
            step_activity_ids = res.fetchall()
            if len(step_activity_ids) == 0:
                raise ValueError("The activity '%s' is not in the database." % activity.value)
            else:
                return tuple()
            
        # -- find task IDs (concrete steps) with that activity
        sql = """
        SELECT DISTINCT step_concrete.id
        FROM step_variant, 
             step_concrete,
             step_concrete2targetfile
        WHERE step_type_id=?
        AND step_variant.id=step_variant_id
        AND step_concrete.id=step_concrete2targetfile.step_concrete_id
        """
        step_concrete_ids = list()
        for step_type_id in step_type_ids:
            res = cursor.execute(sql, step_type_id)
            step_concrete_ids.extend(res.fetchall())

        # -- find the stored entities created by the task above
        results = list()
        for task_dbentry in step_concrete_ids:
            res = self.get_targetassets(task_dbentry[0])
            results.extend(res)
        
        return tuple(results)

    def get_targetsoftype(self, clsname):
        """ Return all targets of a given type. """
        cursor = self.connection.cursor()
        sql = """
        SELECT stored_entity.id,
               classname,
               entityname
        FROM stored_entity
        INNER JOIN step_concrete2targetfile
        ON stored_entity.id=step_concrete2targetfile.stored_entity_id
        WHERE stored_entity.classname=?
        """
        res = cursor.execute(sql, (clsname,))
        results = tuple(self.StoredEntityNoLabel(*x) for x in res.fetchall())
        return results

    def iter_finaltargets(self):
        """ Targets not used as source anywhere else. """
        cursor = self.connection.cursor()
        sql = """
        SELECT se.id, se.classname, se.entityname, step_concrete_id, label
        FROM stored_entity AS se
        INNER JOIN (
          -- select targets not used used as source
          SELECT trg.stored_entity_id, trg.step_concrete_id
          FROM step_concrete2targetfile AS trg
          LEFT JOIN step_concrete2srcfile AS src
          ON trg.stored_entity_id=src.stored_entity_id
          WHERE src.stored_entity_id IS NULL
          )
        ON se.id=stored_entity_id
        INNER JOIN step_concrete
        ON step_concrete.id=step_concrete_id
        INNER JOIN step_status
        ON step_status_id=step_status.id
        ORDER BY se.id
        """
        cursor.execute(sql)
        Target = namedtuple('Target',
                            'stored_entity_id stored_entity_classname stored_entity_entityname step_concrete_id status_label')
        for x in cursor:
            yield Target(*x)

    def finalsteps(self):
        """ Concrete steps for which all targets are final """
        # concrete steps for which all targets are final
        raise NotImplementedError()



class CachedPersistentTaskGraph(PersistentTaskGraph):

    def __init__(self, db_fn, model, wd='.', force_create=False, isolation_level=None):
        super(CachedPersistentTaskGraph, self).__init__(db_fn, model, wd=wd, 
                                                        force_create=force_create,
                                                        isolation_level=isolation_level)
        # cache of DB IDs
        self._cache_steptype_dbid = dict()
        self._cache_task_dbid = dict()
        self._cache_dbid_task_targets = dict()
        self._cache_stepvariant_dbid = dict()
        self._cache_dbid_stepvariant = dict()
        self._cache_stepparameters_dbid = dict()
        self._cache_storedentity_dbid = dict()
        self._cache_storedsequence_dbid = dict()

    def id_step_type(self, activities):
        hashdb = activities
        dbid = self._cache_steptype_dbid.get(hashdb)
        if dbid is not None:
            res = DbID(dbid, False)
        else:
            res = super(CachedPersistentTaskGraph, self).id_step_type(activities)
            self._cache_steptype_dbid[hashdb] = res.id
        return res

    def _cache_id_step_type(self):
        for dbid, activities in itertools.groupby(operator.itemgetter(0),
                                                  self._iter_step_type()):
            hashdb = tuple(activities)
            self._cache_steptype_dbid[hashdb] = dbid

    def id_step_variant(self, step, activities):
        cls = type(step)
        #FIXME: redundancy between the full class name and the activities)
        step_hashdb = (cls.__module__ + '.' + cls.__name__, activities)
        variant_hashdb = (step.version, step._execpath)
        hashdb = (step_hashdb, variant_hashdb)
        dbid = self._cache_stepvariant_dbid.get(hashdb)
        if dbid is not None:
            res = DbID(dbid, False)
        else:
            res = super(CachedPersistentTaskGraph, self).id_step_variant(step, activities)
            self._cache_stepvariant_dbid[hashdb] = res.id
            self._cache_dbid_stepvariant[res.id] = (step_hashdb, variant_hashdb)
        return res


    def id_stepconcrete(self, step_variant_id,
                        sources, targets, parameters,
                        tag = 1):
        """
        Conditionally add a task ("concrete" step),
        that is a step variant (executable and parameters)
        to which source and target files, as well as parameters, are added.

        :param step_variant_id: ID for the step variant
        :type step_variant_id: integer
        :param sources: sequence of sources
        :type sources: :class:`AssetSet`
        :param targets: sequence of targets
        :type targets: :class:`AssetSet`
        :param parameters: list of parameters
        :type parameters: a sequence of :class:`str`
        :param tag: a tag, used to performed repetitions of the exact same task
        :type tag: a sequence of :class:`int`

        :rtype: :class:`DbID`

        """

        assert isinstance(step_variant_id, int)

        step_variant_hashdb = self._cache_dbid_stepvariant.get(step_variant_id)
        if step_variant_hashdb is None:
            raise ValueError('There is no step variant with ID %i' % step_variant_id)
        step_hashdb, variant_hashdb = step_variant_hashdb
        task_hashdb = (step_hashdb, sources.hashdb, parameters, tag)
        dbid = self._cache_task_dbid.get(task_hashdb)
        if dbid is not None:

            res = DbID(dbid, False)

            # Some of the targets might be undefined (because the user leaves it to railroadtracks to
            # fill the blanks). We need to populate these with the values in the database in order to
            # be able to compute the hash.
            if any(not getattr(targets, x.label)._defined for x in self.get_targetassets(res)):
                # FIXME: for now just hit the database (no use of the cache)
                res = super(CachedPersistentTaskGraph, self).id_stepconcrete(step_variant_id, 
                                                                             sources, 
                                                                             targets,
                                                                             parameters, 
                                                                             tag=tag)
            elif dbid in self._cache_dbid_task_targets:
                # check that the target assets have not changed
                if targets.hashdb != self._cache_dbid_task_targets[dbid]:
                    raise ValueError("Target assets not matching the assets already associated with the task.")
            else:
                self._cache_dbid_task_targets[res.id] = targets.hashdb
            
        else:
            res = super(CachedPersistentTaskGraph, self).id_stepconcrete(step_variant_id, sources, targets, parameters, tag=tag)
            self._cache_task_dbid[task_hashdb] = res.id
            if all(x._defined for x in targets):
                self._cache_dbid_task_targets[res.id] = targets.hashdb
        return res


    def id_stepparameters(self, parameters):
        
        dbid = self._cache_stepparameters_dbid.get(parameters)
        if dbid is not None:
            res = DbID(dbid, False)
        else:
            res = super(CachedPersistentTaskGraph, self).id_stepparameters(parameters)
            self._cache_stepparameters_dbid[parameters] = res.id
        return res


    def id_stored_entity(self, cls, name):
        hashdb = core.SavedEntityAbstract._hash_components(cls, name)
        dbid = self._cache_storedentity_dbid.get(hashdb)
        if dbid is not None:
            res = DbID(dbid, False)
        else:
            res = super(CachedPersistentTaskGraph, self).id_stored_entity(cls, name)
            self._cache_storedentity_dbid[hashdb] = res.id
        return res
        

    def id_stored_sequence(self, cls, clsname_sequence):
        names_hashdb = list()
        clsname_sequence_tpl = tuple(clsname_sequence)
        for (elt_cls, elt_name) in clsname_sequence_tpl:
            dbid = self.id_stored_entity(elt_cls, elt_name) 
            item_hashdb = core.SavedEntityAbstract._hash_components(elt_cls, elt_name)
            names_hashdb.append(item_hashdb)
        hashdb = (cls, tuple(names_hashdb))
        dbid = self._cache_storedsequence_dbid.get(hashdb)
        if dbid is not None:
            res = DbID(dbid, False)
        else:
            res = super(CachedPersistentTaskGraph, self).id_stored_sequence(cls, clsname_sequence_tpl)
            self._cache_storedsequence_dbid[hashdb] = res.id
        return res
        



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
import tempfile, os, time, sys
import subprocess
import multiprocessing

import abc

from railroadtracks import hortator
from railroadtracks.easy.tasks import TaskSet, DbID

from six import with_metaclass

class ParallelExecutionAbstract(with_metaclass(abc.ABCMeta, object)):

    def map(self, taskset):
        raise NotImplementedError()
    def map_async(self, taskset):
        raise NotImplementedError()


def _execute_subprocess_in_dir(x):
    # expand 'x'
    taskid, cmd, wd = x
    import subprocess, tempfile, os, time
    od = os.path.abspath(os.curdir)
    t0 = None
    t1 = None
    env = os.environ.copy()
    if 'PYTHONPATH' in env:
        pythonpath = os.getenv('PYTHONPATH')
        env['PYTHONPATH'] = ':'.join((od, pythonpath))
    else:
        env['PYTHONPATH'] = od
    #env = {}
    try:
        os.chdir(wd)
        t0 = time.time()
        returncode = subprocess.check_call(cmd,
                                           env = env)
        t1 = time.time()
        return (taskid, returncode, t0, t1)
    except:
        return (taskid, 1, t0, t1)
    finally:
        os.chdir(od)

def _set_task_status(project, res):
    """
    :param res: iterable of (taskid, returncode) tuples.
    """
    ct = Counter()
    for x in res:
        try:
            taskid, returncode, t0, t1 = x
        except:
            raise ValueError('Unable to unpack values from:\n%s' % str(x))
        if returncode == 0:
            status = hortator._TASK_DONE
        else:
            status = hortator._TASK_FAILED
        if not hasattr(taskid, 'id'):
            taskid = DbID(taskid)
        project.persistent_graph.step_concrete_state(taskid,
                                                     hortator._TASK_STATUS_LIST[status])
        project.persistent_graph.set_time_t0(taskid, t0)
        project.persistent_graph.set_time_t1(taskid, t1)
        ct.update((status,))
    return ct

class DummyExecution(ParallelExecutionAbstract):
    def __init__(self):
        pass

    def map(self, taskset, tempdir=None):
        ct = Counter()
        for task in taskset:
            ct.update(task.status)
        return ct

class IterativeExecution(ParallelExecutionAbstract):
    """ Task mapper executing tasks in sequence and within the current process. """
    def __init__(self):
        pass

    def map(self, taskset, tempdir=None):
        assert isinstance(taskset, TaskSet)
        if len(taskset) == 0:
            return
        project = next(iter(taskset)).project
        od = os.curdir
        ct = Counter()
        for task in taskset:
            t1 = None
            wd = tempfile.mkdtemp(dir=tempdir)
            os.chdir(wd)
            t0 = time.time()
            try:
                task.execute()
                t1 = time.time()
                returncode = 0
            except:
                returncode = 1
            finally:
                pass
                os.chdir(od)
                res = (task.task_id, returncode, t0, t1)
                ct.update(_set_task_status(project, (res,)))

        return ct


class MultiprocessingExecution(ParallelExecutionAbstract):
    """ Task mapper using Python's multiprocessing """
    def __init__(self, processes, python_executable=sys.executable):
        """ 
        :param processes: number of simultaneous Python processing (underlying tools may use threads or processes themselves)
        """
        self.__processes = processes
        self._pyexecutable = python_executable

    def __get_ressources(self):
        self._pool = multiprocessing.Pool(self.__processes)
        
    def map(self, taskset, tempdir=None):
        assert isinstance(taskset, TaskSet)
        if len(taskset) == 0:
            return
        self.__get_ressources()
        pool = self._pool
        wds = (tempfile.mkdtemp(dir=tempdir) for x in taskset)
        ct = None
        res = None
        try:
            gen = ((x.task_id, x.unifex_cmd(python_executable=self._pyexecutable), y)
                   for x,y 
                   in zip(taskset, wds))
            res = pool.map(_execute_subprocess_in_dir, 
                           gen)
        finally:
            pool.close()
            pool.join()
            for d in wds:
                shutil.rmtree(d)
        if res is not None:
            project = next(iter(taskset)).project
            ct = _set_task_status(project, res)
        return ct

    def __str__(self):
        return "%s using %i processes." % (str(type(self)), self.__processes)


def select_mapper(taskset, mapper_smallset, mapper_largeset):
    if len(taskset) < 3:
        res = mapper_smallset
    else:
        res = mapper_largeset
    return res

class DuplexExecution(ParallelExecutionAbstract):
    """ """
    def __init__(self, mapper_a, mapper_b, func):
        self._mapper_a = mapper_a
        self._mapper_b = mapper_b
        self._func = func
    def map(self, taskset, tempdir=None):
        mapper = self._func(taskset, self._mapper_a, self._mapper_b)
        return mapper.map(taskset)

def exec_taskfilter(taskset):
    """ Returns a taskset in which tasks are candidates for execution. """
    # union of tasks that are either "TO-DO" or "FAILED"
    tasks_todo = taskset.filter_on_status(hortator._TASK_TODO)
    tasks_todo = tasks_todo.union(taskset.filter_on_status(hortator._TASK_FAILED))
    # only keep the tasks for which the parent task was successfully run
    tasks_todo = tasks_todo.filter_on_parent_status(hortator._TASK_DONE)
    return tasks_todo
